@@ -1,9 +1,9 @@
-#!/usr/bin/python3
+#!/usr/bin/env python3
 # pylint: disable=invalid-name,missing-docstring
 #
-# Copyright (C) 2021 Richard Hughes <richard@hughsie.com>
+# Copyright 2021 Richard Hughes <richard@hughsie.com>
 #
-# SPDX-License-Identifier: LGPL-2.1+
+# SPDX-License-Identifier: LGPL-2.1-or-later
 #
 # pylint: disable=too-many-instance-attributes,no-self-use
 
@@ -18,7 +18,6 @@ DEFAULT_BUILDDIR = ".ossfuzz"
 
 class Builder:
     def __init__(self) -> None:
-
         self.cc = self._ensure_environ("CC", "gcc")
         self.cxx = self._ensure_environ("CXX", "g++")
         self.builddir = self._ensure_environ("WORK", os.path.realpath(DEFAULT_BUILDDIR))
@@ -33,6 +32,7 @@ class Builder:
             "-lffi",
             "-lz",
             "-llzma",
+            "-lzstd",
         ]
 
         # defined in env
@@ -63,7 +63,13 @@ class Builder:
             os.environ[key] = value
         return os.environ[key]
 
-    def checkout_source(self, name: str, url: str, commit: Optional[str] = None) -> str:
+    def checkout_source(
+        self,
+        name: str,
+        url: str,
+        commit: Optional[str] = None,
+        patches: Optional[List[str]] = None,
+    ) -> str:
         """checkout source tree, optionally to a specific commit"""
         srcdir_name = os.path.join(self.srcdir, name)
         if os.path.exists(srcdir_name):
@@ -71,6 +77,12 @@ class Builder:
         subprocess.run(["git", "clone", url], cwd=self.srcdir, check=True)
         if commit:
             subprocess.run(["git", "checkout", commit], cwd=srcdir_name, check=True)
+        if patches:
+            for fn in patches:
+                with open(os.path.join(self.srcdir, "fwupd", fn), "rb") as f:
+                    subprocess.run(
+                        ["patch", "-p1"], cwd=srcdir_name, check=True, input=f.read()
+                    )
         return srcdir_name
 
     def build_meson_project(self, srcdir: str, argv) -> None:
@@ -104,8 +116,8 @@ class Builder:
             subprocess.run(
                 [
                     "cmake",
-                    "-DCMAKE_INSTALL_PREFIX:PATH={}".format(self.builddir),
-                    "-DCMAKE_INSTALL_LIBDIR={}".format("lib"),
+                    f"-DCMAKE_INSTALL_PREFIX:PATH={self.builddir}",
+                    "-DCMAKE_INSTALL_LIBDIR=lib",
                 ]
                 + argv
                 + [".."],
@@ -116,11 +128,11 @@ class Builder:
 
     def add_work_includedir(self, value: str) -> None:
         """add a CFLAG"""
-        self.cflags.append("-I{}/{}".format(self.builddir, value))
+        self.cflags.append(f"-I{self.builddir}/{value}")
 
     def add_src_includedir(self, value: str) -> None:
         """add a CFLAG"""
-        self.cflags.append("-I{}/{}".format(self.srcdir, value))
+        self.cflags.append(f"-I{self.srcdir}/{value}")
 
     def add_build_ldflag(self, value: str) -> None:
         """add a LDFLAG"""
@@ -130,7 +142,7 @@ class Builder:
         """map changes"""
 
         dst = os.path.basename(src).replace(".in", "")
-        with open(os.path.join(self.srcdir, src), "r") as f:
+        with open(os.path.join(self.srcdir, src)) as f:
             blob = f.read()
             for key in replacements:
                 blob = blob.replace(key, replacements[key])
@@ -147,16 +159,37 @@ class Builder:
             fullsrc = os.path.join(self.builddir, src)
         dst = os.path.basename(src).replace(".c", ".o")
         argv.extend(["-c", fullsrc, "-o", os.path.join(self.builddir, dst)])
-        print("building {} into {}".format(src, dst))
+        print(f"building {src} into {dst}")
         try:
             subprocess.run(argv, cwd=self.srcdir, check=True)
         except subprocess.CalledProcessError as e:
             print(e)
             sys.exit(1)
-        return os.path.join(self.builddir, "{}".format(dst))
+        return os.path.join(self.builddir, f"{dst}")
 
-    def link(self, objs: List[str], dst: str) -> None:
-        """link multiple obects into a binary"""
+    def rustgen(self, src: str) -> str:
+        fn_root = os.path.basename(src).replace(".rs", "")
+        fulldst_c = os.path.join(self.builddir, f"{fn_root}-struct.c")
+        fulldst_h = os.path.join(self.builddir, f"{fn_root}-struct.h")
+        try:
+            subprocess.run(
+                [
+                    "python",
+                    "fwupd/libfwupdplugin/rustgen.py",
+                    src,
+                    fulldst_c,
+                    fulldst_h,
+                ],
+                cwd=self.srcdir,
+                check=True,
+            )
+        except subprocess.CalledProcessError as e:
+            print(e)
+            sys.exit(1)
+        return fulldst_c
+
+    def link(self, objs: List[str], dst: str) -> str:
+        """link multiple objects into a binary"""
         argv = [self.cxx] + self.cxxflags
         for obj in objs:
             if obj.startswith("-"):
@@ -165,31 +198,29 @@ class Builder:
                 argv.append(os.path.join(self.builddir, obj))
         argv += ["-o", os.path.join(self.installdir, dst)]
         argv += self.ldflags
-        print("building {} into {}".format(",".join(objs), dst))
+        print(f"building {','.join(objs)} into {dst}")
         subprocess.run(argv, cwd=self.srcdir, check=True)
+        return os.path.join(self.installdir, dst)
 
-    def mkfuzztargets(self, globstr: str) -> None:
+    def mkfuzztargets(self, exe: str, globstr: str) -> List[str]:
         """make binary fuzzing targets from builder.xml files"""
         builder_xmls = glob.glob(globstr)
+        corpus: List[str] = []
         if not builder_xmls:
-            print("failed to find {}".format(globstr))
-            sys.exit(1)
+            print(f"failed to find {globstr}")
         for fn_src in builder_xmls:
-            fn_dst = fn_src.replace(".builder.xml", ".bin")
-            if os.path.exists(fn_dst):
-                continue
-            print("building {} into {}".format(fn_src, fn_dst))
+            fn_dst = os.path.join(
+                self.builddir, os.path.basename(fn_src).replace(".builder.xml", ".bin")
+            )
+            print(f"building {fn_src} into {fn_dst}")
             try:
-                argv = [
-                    "build/src/fwupdtool",
-                    "firmware-build",
-                    fn_src,
-                    fn_dst,
-                ]
+                argv = [exe, fn_src, fn_dst]
                 subprocess.run(argv, check=True)
             except subprocess.CalledProcessError as e:
-                print("tried to run: `{}` and got {}".format(" ".join(argv), str(e)))
+                print(f"tried to run: `{' '.join(argv)}` and got {str(e)}")
                 sys.exit(1)
+            corpus.append(fn_dst)
+        return corpus
 
     def write_header(
         self, dst: str, defines: Dict[str, Optional[Union[str, int]]]
@@ -197,31 +228,31 @@ class Builder:
         """write a header file"""
         dstdir = os.path.join(self.builddir, os.path.dirname(dst))
         os.makedirs(dstdir, exist_ok=True)
-        print("writing {}".format(dst))
+        print(f"writing {dst}")
         with open(os.path.join(dstdir, os.path.basename(dst)), "w") as f:
             for key in defines:
                 value = defines[key]
                 if value is not None:
                     if isinstance(value, int):
-                        f.write("#define {} {}\n".format(key, value))
+                        f.write(f"#define {key} {value}\n")
                     else:
-                        f.write('#define {} "{}"\n'.format(key, value))
+                        f.write(f'#define {key} "{value}"\n')
                 else:
-                    f.write("#define {}\n".format(key))
+                    f.write(f"#define {key}\n")
         self.add_work_includedir(os.path.dirname(dst))
 
-    def makezip(self, dst: str, globstr: str) -> None:
+    def makezip(self, dst: str, corpus: List[str]) -> None:
         """create a zip file archive from a glob"""
-        argv = ["zip", "--junk-paths", os.path.join(self.installdir, dst)] + glob.glob(
-            os.path.join(self.srcdir, globstr)
-        )
-        print("assembling {}".format(dst))
+        if not corpus:
+            return
+        argv = ["zip", "--junk-paths", os.path.join(self.installdir, dst)] + corpus
+        print(f"assembling {dst}")
         subprocess.run(argv, cwd=self.srcdir, check=True)
 
     def grep_meson(self, src: str, token: str = "fuzzing") -> List[str]:
         """find source files tagged with a specific comment"""
         srcs = []
-        with open(os.path.join(self.srcdir, src, "meson.build"), "r") as f:
+        with open(os.path.join(self.srcdir, src, "meson.build")) as f:
             for line in f.read().split("\n"):
                 if line.find(token) == -1:
                     continue
@@ -245,29 +276,29 @@ class Builder:
 
 
 class Fuzzer:
-    def __init__(self, name, srcdir=None, globstr=None, pattern=None) -> None:
-
+    def __init__(self, name, srcdir=None, pattern=None) -> None:
         self.name = name
         self.srcdir = srcdir or name
-        self.globstr = globstr or "{}*.bin".format(name)
-        self.pattern = pattern or "{}-firmware".format(name)
+        self.globstr = f"{name}*.bin"
+        self.pattern = pattern or f"{name}-firmware"
 
     @property
     def new_gtype(self) -> str:
-        return "fu_{}_new".format(self.pattern).replace("-", "_")
+        return f"g_object_new(FU_TYPE_{self.pattern.replace('-', '_').upper()}, NULL)"
 
     @property
     def header(self) -> str:
-        return "fu-{}.h".format(self.pattern)
+        return f"fu-{self.pattern}.h"
 
 
 def _build(bld: Builder) -> None:
-
     # CBOR
     src = bld.checkout_source(
-        "libcbor", url="https://github.com/PJK/libcbor.git", commit="v0.9.0"
+        "libcbor",
+        url="https://github.com/PJK/libcbor.git",
+        commit="b223daaaa34dcb83f9c25576f05e4f1646f44bf9",
     )
-    bld.build_cmake_project(src)
+    bld.build_cmake_project(src, argv=["-DCMAKE_INTERPROCEDURAL_OPTIMIZATION=OFF"])
     bld.add_build_ldflag("lib/libcbor.a")
 
     # GLib
@@ -297,7 +328,9 @@ def _build(bld: Builder) -> None:
 
     # JSON-GLib
     src = bld.checkout_source(
-        "json-glib", url="https://gitlab.gnome.org/GNOME/json-glib.git"
+        "json-glib",
+        url="https://gitlab.gnome.org/GNOME/json-glib.git",
+        commit="1.8.0-actual",
     )
     bld.build_meson_project(
         src, ["-Dgtk_doc=disabled", "-Dtests=false", "-Dintrospection=disabled"]
@@ -316,48 +349,61 @@ def _build(bld: Builder) -> None:
     bld.add_build_ldflag("lib/libxmlb.a")
 
     # write required headers
-    bld.write_header("libfwupd/fwupd-version.h", {})
+    bld.write_header(
+        "libfwupd/fwupd-version.h",
+        {
+            "FWUPD_MAJOR_VERSION": 0,
+            "FWUPD_MINOR_VERSION": 0,
+            "FWUPD_MICRO_VERSION": 0,
+        },
+    )
     bld.write_header(
         "config.h",
         {
             "FWUPD_DATADIR": "/tmp",
+            "FWUPD_DATADIR_VENDOR_IDS": "/tmp",
             "FWUPD_LOCALSTATEDIR": "/tmp",
             "FWUPD_LIBDIR_PKG": "/tmp",
             "FWUPD_SYSCONFDIR": "/tmp",
+            "FWUPD_LIBEXECDIR": "/tmp",
+            "HAVE_CBOR": None,
+            "HAVE_CBOR_SET_ALLOCS": None,
             "HAVE_REALPATH": None,
             "PACKAGE_NAME": "fwupd",
             "PACKAGE_VERSION": "0.0.0",
         },
     )
-    bld.write_header(
-        "libfwupdplugin/fu-version.h",
-        {
-            "FU_MAJOR_VERSION": 0,
-            "FU_MINOR_VERSION": 0,
-            "FU_MICRO_VERSION": 0,
-        },
-    )
 
     # libfwupd + libfwupdplugin
     built_objs: List[str] = []
+    fuzzing_objs: List[str] = []
     bld.add_src_includedir("fwupd")
     for path in ["fwupd/libfwupd", "fwupd/libfwupdplugin"]:
         bld.add_src_includedir(path)
         for src in bld.grep_meson(path):
-            built_objs.append(bld.compile(src))
+            if src.endswith(".c"):
+                built_objs.append(bld.compile(src))
+            elif src.endswith(".rs"):
+                built_objs.append(bld.compile(bld.rustgen(src)))
 
     # dummy binary entrypoint
     if "LIB_FUZZING_ENGINE" in os.environ:
-        built_objs.append(os.environ["LIB_FUZZING_ENGINE"])
+        fuzzing_objs.append(os.environ["LIB_FUZZING_ENGINE"])
     else:
-        built_objs.append(bld.compile("fwupd/libfwupdplugin/fu-fuzzer-main.c"))
+        fuzzing_objs.append(bld.compile("fwupd/libfwupdplugin/fu-fuzzer-main.c"))
 
     # built in formats
     for fzr in [
+        Fuzzer("efi-lz77", pattern="efi-lz77-decompressor"),
+        Fuzzer("csv"),
+        Fuzzer("cab"),
         Fuzzer("dfuse"),
+        Fuzzer("edid", pattern="edid"),
+        Fuzzer("elf"),
         Fuzzer("fdt"),
         Fuzzer("fit"),
         Fuzzer("fmap"),
+        Fuzzer("hid-descriptor", pattern="hid-descriptor"),
         Fuzzer("ihex"),
         Fuzzer("srec"),
         Fuzzer("intel-thunderbolt"),
@@ -365,8 +411,9 @@ def _build(bld: Builder) -> None:
         Fuzzer("ifwi-fpt"),
         Fuzzer("oprom"),
         Fuzzer("uswid"),
-        Fuzzer("efi-firmware-filesystem", pattern="efi-firmware-filesystem"),
-        Fuzzer("efi-firmware-volume", pattern="efi-firmware-volume"),
+        Fuzzer("efi-filesystem", pattern="efi-filesystem"),
+        Fuzzer("efi-volume", pattern="efi-volume"),
+        Fuzzer("efi-load-option", pattern="efi-load-option"),
         Fuzzer("ifd"),
     ]:
         src = bld.substitute(
@@ -376,27 +423,41 @@ def _build(bld: Builder) -> None:
                 "@INCLUDE@": os.path.join("libfwupdplugin", fzr.header),
             },
         )
-        bld.link([bld.compile(src)] + built_objs, "{}_fuzzer".format(fzr.name))
-        bld.mkfuzztargets(
+        exe = bld.link(
+            [bld.compile(src)] + fuzzing_objs + built_objs, f"{fzr.name}_fuzzer"
+        )
+
+        src_generator = bld.substitute(
+            "fwupd/libfwupdplugin/fu-fuzzer-generate.c.in",
+            {
+                "@FIRMWARENEW@": fzr.new_gtype,
+                "@INCLUDE@": os.path.join("libfwupdplugin", fzr.header),
+            },
+        )
+        exe_generator = bld.link(
+            [bld.compile(src_generator)] + built_objs, f"{fzr.name}_generator"
+        )
+        corpus = bld.mkfuzztargets(
+            exe_generator,
             os.path.join(
                 bld.srcdir,
                 "fwupd",
                 "libfwupdplugin",
                 "tests",
-                "{}*.builder.xml".format(fzr.name),
-            )
+                f"{fzr.name}*.builder.xml",
+            ),
         )
         bld.makezip(
-            "{}_fuzzer_seed_corpus.zip".format(fzr.name),
-            "fwupd/libfwupdplugin/tests/{}".format(fzr.globstr),
+            f"{fzr.name}_fuzzer_seed_corpus.zip",
+            corpus,
         )
 
     # plugins
     for fzr in [
         Fuzzer("acpi-phat", pattern="acpi-phat"),
         Fuzzer("bcm57xx"),
-        Fuzzer("ccgx-dmc", srcdir="ccgx", globstr="ccgx-dmc*.bin"),
         Fuzzer("ccgx"),
+        Fuzzer("ccgx-dmc"),
         Fuzzer("cros-ec"),
         Fuzzer("ebitdo"),
         Fuzzer("elanfp"),
@@ -413,8 +474,11 @@ def _build(bld: Builder) -> None:
         Fuzzer("wacom-usb", pattern="wac-firmware"),
     ]:
         fuzz_objs = []
-        for obj in bld.grep_meson("fwupd/plugins/{}".format(fzr.srcdir)):
-            fuzz_objs.append(bld.compile(obj))
+        for obj in bld.grep_meson(f"fwupd/plugins/{fzr.srcdir}"):
+            if obj.endswith(".c"):
+                fuzz_objs.append(bld.compile(obj))
+            elif obj.endswith(".rs"):
+                fuzz_objs.append(bld.compile(bld.rustgen(obj)))
         src = bld.substitute(
             "fwupd/libfwupdplugin/fu-fuzzer-firmware.c.in",
             {
@@ -422,30 +486,55 @@ def _build(bld: Builder) -> None:
                 "@INCLUDE@": os.path.join("plugins", fzr.srcdir, fzr.header),
             },
         )
-        fuzz_objs.append(bld.compile(src))
-        bld.link(fuzz_objs + built_objs, "{}_fuzzer".format(fzr.name))
-        bld.mkfuzztargets(
+        exe = bld.link(
+            fuzz_objs + built_objs + fuzzing_objs + [bld.compile(src)],
+            f"{fzr.name}_fuzzer",
+        )
+
+        # generate the corpus
+        src_generator = bld.substitute(
+            "fwupd/libfwupdplugin/fu-fuzzer-generate.c.in",
+            {
+                "@FIRMWARENEW@": fzr.new_gtype,
+                "@INCLUDE@": os.path.join("plugins", fzr.srcdir, fzr.header),
+            },
+        )
+        exe_generator = bld.link(
+            fuzz_objs + built_objs + [bld.compile(src_generator)],
+            f"{fzr.name}_generator",
+        )
+        corpus = bld.mkfuzztargets(
+            exe_generator,
             os.path.join(
                 bld.srcdir,
                 "fwupd",
                 "plugins",
                 fzr.srcdir,
                 "tests",
-                "{}*.builder.xml".format(fzr.name),
-            )
+                f"{fzr.name}*.builder.xml",
+            ),
         )
         bld.makezip(
-            "{}_fuzzer_seed_corpus.zip".format(fzr.name),
-            "fwupd/plugins/{}/tests/{}".format(fzr.srcdir, fzr.globstr),
+            f"{fzr.name}_fuzzer_seed_corpus.zip",
+            corpus,
         )
 
 
 if __name__ == "__main__":
-
     # install missing deps here rather than patching the Dockerfile in oss-fuzz
     try:
         subprocess.check_call(
-            ["apt-get", "install", "-y", "liblzma-dev", "libcbor-dev"],
+            [
+                "apt-get",
+                "install",
+                "-y",
+                "liblzma-dev",
+                "libzstd-dev",
+                "libcbor-dev",
+                "python3",
+                "python3-jinja2",
+                "python3-packaging",
+            ],
             stdout=open(os.devnull, "wb"),
         )
     except FileNotFoundError:
