@@ -30,11 +30,27 @@ LDFLAGS=$(dpkg-buildflags --get LDFLAGS | sed "s/-Wl,-Bsymbolic-functions\s//")
 export LDFLAGS
 
 root=$(pwd)
-rm -rf ${root}/build
+export BUILD=${root}/build
+rm -rf ${BUILD}
 chown -R nobody ${root}
-sudo -u nobody meson ${root}/build -Dman=false -Ddocs=enabled -Dgusb:tests=false --prefix=${root}/dist
+sudo -u nobody meson ${BUILD}               \
+                    -Dman=false             \
+                    -Ddocs=enabled          \
+                    -Dlibxmlb:gtkdoc=false  \
+                    --prefix=${root}/dist
 #build with clang
-sudo -u nobody ninja -C ${root}/build test -v
+sudo -u nobody ninja -C ${BUILD} test -v
+
+# check for unused symbols
+./contrib/ci/check-unused.py
+
+# check the daemon aborts
+set +e
+FWUPD_SYSCALL_FILTER=systemd ${BUILD}/src/fwupd --immediate-exit
+if [ $? -ne 1 ] ; then
+    echo "failed to detect missing syscall filtering"
+    exit 1
+fi
 
 #make docs available outside of docker
-ninja -C ${root}/build install -v
+ninja -C ${BUILD} install -v
