@@ -1,17 +1,17 @@
 // Copyright 2024 Richard Hughes <richard@hughsie.com>
 // SPDX-License-Identifier: LGPL-2.1-or-later
 
-#[derive(ValidateBytes)]
-struct ParadeUsbhubHdr {
+#[derive(ValidateStream)]
+struct FuStructParadeUsbhubHdr {
     signature: u16be == 0x55AA,
 }
 
-enum ParadeUsbhubDeviceRequest {
+enum FuParadeUsbhubDeviceRequest {
     Read = 0x40,
     Write = 0x41,
 }
 
-enum ParadeUsbhubDeviceAddr {
+enum FuParadeUsbhubDeviceAddr {
     Status      = 0x5000,
     Data        = 0x5001, // u32
     SpiAddr     = 0x5005, // u24
@@ -29,7 +29,7 @@ enum ParadeUsbhubDeviceAddr {
     VersionD    = 0x5C12,
 }
 
-enum ParadeUsbhubDeviceStatusFlag {
+enum FuParadeUsbhubDeviceStatusFlag {
     Write       = 0b00000001,
     TriggerSpi  = 0b00000010,
     TriggerDbi  = 0b00000100,
