@@ -1,13 +1,11 @@
 /*
- * Copyright (C) 2018 Richard Hughes <richard@hughsie.com>
+ * Copyright 2018 Richard Hughes <richard@hughsie.com>
  *
- * SPDX-License-Identifier: LGPL-2.1+
+ * SPDX-License-Identifier: LGPL-2.1-or-later
  */
 
 #include "config.h"
 
-#include <fwupdplugin.h>
-
 #include "fu-rts54hid-common.h"
 #include "fu-rts54hid-device.h"
 
@@ -23,8 +21,8 @@ static void
 fu_rts54hid_device_to_string(FuDevice *device, guint idt, GString *str)
 {
 	FuRts54HidDevice *self = FU_RTS54HID_DEVICE(device);
-	fu_string_append_kb(str, idt, "FwAuth", self->fw_auth);
-	fu_string_append_kb(str, idt, "DualBank", self->dual_bank);
+	fwupd_codec_string_append_bool(str, idt, "FwAuth", self->fw_auth);
+	fwupd_codec_string_append_bool(str, idt, "DualBank", self->dual_bank);
 }
 
 static gboolean
@@ -182,7 +180,7 @@ fu_rts54hid_device_verify_update_fw(FuRts54HidDevice *self, FuProgress *progress
 				      FU_HID_DEVICE_FLAG_NONE,
 				      error))
 		return FALSE;
-	fu_progress_sleep(progress, 4000);
+	fu_device_sleep_full(FU_DEVICE(self), 4000, progress); /* ms */
 	if (!fu_hid_device_get_report(FU_HID_DEVICE(self),
 				      0x0,
 				      buf,
@@ -339,8 +337,8 @@ fu_rts54hid_device_write_firmware(FuDevice *device,
 				  GError **error)
 {
 	FuRts54HidDevice *self = FU_RTS54HID_DEVICE(device);
-	g_autoptr(GBytes) fw = NULL;
-	g_autoptr(GPtrArray) chunks = NULL;
+	g_autoptr(GInputStream) stream = NULL;
+	g_autoptr(FuChunkArray) chunks = NULL;
 
 	/* progress */
 	fu_progress_set_id(progress, G_STRLOC);
@@ -350,8 +348,8 @@ fu_rts54hid_device_write_firmware(FuDevice *device,
 	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_BUSY, 1, "reset");
 
 	/* get default image */
-	fw = fu_firmware_get_bytes(firmware, error);
-	if (fw == NULL)
+	stream = fu_firmware_get_stream(firmware, error);
+	if (stream == NULL)
 		return FALSE;
 
 	/* set MCU to high clock rate for better ISP performance */
@@ -364,12 +362,18 @@ fu_rts54hid_device_write_firmware(FuDevice *device,
 	fu_progress_step_done(progress);
 
 	/* write each block */
-	chunks = fu_chunk_array_new_from_bytes(fw,
-					       0x00, /* start addr */
-					       0x00, /* page_sz */
-					       FU_RTS54HID_TRANSFER_BLOCK_SIZE);
-	for (guint i = 0; i < chunks->len; i++) {
-		FuChunk *chk = g_ptr_array_index(chunks, i);
+	chunks =
+	    fu_chunk_array_new_from_stream(stream, 0x00, FU_RTS54HID_TRANSFER_BLOCK_SIZE, error);
+	if (chunks == NULL)
+		return FALSE;
+	for (guint i = 0; i < fu_chunk_array_length(chunks); i++) {
+		g_autoptr(FuChunk) chk = NULL;
+
+		/* prepare chunk */
+		chk = fu_chunk_array_index(chunks, i, error);
+		if (chk == NULL)
+			return FALSE;
+
 		/* write chunk */
 		if (!fu_rts54hid_device_write_flash(self,
 						    fu_chunk_get_address(chk),
@@ -381,7 +385,7 @@ fu_rts54hid_device_write_firmware(FuDevice *device,
 		/* update progress */
 		fu_progress_set_percentage_full(fu_progress_get_child(progress),
 						(gsize)i + 1,
-						(gsize)chunks->len);
+						(gsize)fu_chunk_array_length(chunks));
 	}
 	fu_progress_step_done(progress);
 
@@ -421,10 +425,10 @@ fu_rts54hid_device_init(FuRts54HidDevice *self)
 static void
 fu_rts54hid_device_class_init(FuRts54HidDeviceClass *klass)
 {
-	FuDeviceClass *klass_device = FU_DEVICE_CLASS(klass);
-	klass_device->write_firmware = fu_rts54hid_device_write_firmware;
-	klass_device->to_string = fu_rts54hid_device_to_string;
-	klass_device->setup = fu_rts54hid_device_setup;
-	klass_device->close = fu_rts54hid_device_close;
-	klass_device->set_progress = fu_rts54hid_device_set_progress;
+	FuDeviceClass *device_class = FU_DEVICE_CLASS(klass);
+	device_class->write_firmware = fu_rts54hid_device_write_firmware;
+	device_class->to_string = fu_rts54hid_device_to_string;
+	device_class->setup = fu_rts54hid_device_setup;
+	device_class->close = fu_rts54hid_device_close;
+	device_class->set_progress = fu_rts54hid_device_set_progress;
 }
