@@ -1,7 +1,7 @@
 /*
- * Copyright (C) 2022 Gaël PORTAY <gael.portay@collabora.com>
+ * Copyright 2022 Gaël PORTAY <gael.portay@collabora.com>
  *
- * SPDX-License-Identifier: LGPL-2.1+
+ * SPDX-License-Identifier: LGPL-2.1-or-later
  */
 
 #include "config.h"
@@ -100,8 +100,7 @@ fu_steelseries_sonic_wireless_status(FuDevice *device,
 				    error))
 		return FALSE;
 
-	if (g_getenv("FWUPD_STEELSERIES_SONIC_VERBOSE") != NULL)
-		fu_dump_raw(G_LOG_DOMAIN, "WirelessStatus", data, sizeof(data));
+	fu_dump_raw(G_LOG_DOMAIN, "WirelessStatus", data, sizeof(data));
 	if (!fu_steelseries_device_cmd(FU_STEELSERIES_DEVICE(device),
 				       data,
 				       sizeof(data),
@@ -109,8 +108,7 @@ fu_steelseries_sonic_wireless_status(FuDevice *device,
 				       error))
 		return FALSE;
 
-	if (g_getenv("FWUPD_STEELSERIES_SONIC_VERBOSE") != NULL)
-		fu_dump_raw(G_LOG_DOMAIN, "WirelessStatus", data, sizeof(data));
+	fu_dump_raw(G_LOG_DOMAIN, "WirelessStatus", data, sizeof(data));
 	if (!fu_memread_uint8_safe(data,
 				   sizeof(data),
 				   STEELSERIES_SONIC_WIRELESS_STATUS_VALUE_OFFSET,
@@ -144,8 +142,7 @@ fu_steelseries_sonic_battery_state(FuDevice *device, guint16 *value, GError **er
 				    error))
 		return FALSE;
 
-	if (g_getenv("FWUPD_STEELSERIES_SONIC_VERBOSE") != NULL)
-		fu_dump_raw(G_LOG_DOMAIN, "BatteryState", data, sizeof(data));
+	fu_dump_raw(G_LOG_DOMAIN, "BatteryState", data, sizeof(data));
 	if (!fu_steelseries_device_cmd(FU_STEELSERIES_DEVICE(device),
 				       data,
 				       sizeof(data),
@@ -153,8 +150,7 @@ fu_steelseries_sonic_battery_state(FuDevice *device, guint16 *value, GError **er
 				       error))
 		return FALSE;
 
-	if (g_getenv("FWUPD_STEELSERIES_SONIC_VERBOSE") != NULL)
-		fu_dump_raw(G_LOG_DOMAIN, "BatteryState", data, sizeof(data));
+	fu_dump_raw(G_LOG_DOMAIN, "BatteryState", data, sizeof(data));
 	if (!fu_memread_uint16_safe(data,
 				    sizeof(data),
 				    STEELSERIES_SONIC_BATTERY_VALUE_OFFSET,
@@ -219,8 +215,7 @@ fu_steelseries_sonic_read_from_ram(FuDevice *device,
 					       TRUE,
 					       error))
 			return FALSE;
-		if (g_getenv("FWUPD_STEELSERIES_SONIC_VERBOSE") != NULL)
-			fu_dump_raw(G_LOG_DOMAIN, "ReadFromRAM", data, sizeof(data));
+		fu_dump_raw(G_LOG_DOMAIN, "ReadFromRAM", data, sizeof(data));
 
 		if (!fu_memcpy_safe(fu_chunk_get_data_out(chk),
 				    fu_chunk_get_data_sz(chk),
@@ -244,7 +239,7 @@ fu_steelseries_sonic_read_from_flash(FuDevice *device,
 				     SteelseriesSonicChip chip,
 				     guint32 address,
 				     guint8 *buf,
-				     guint32 bufsz,
+				     gsize bufsz,
 				     FuProgress *progress,
 				     GError **error)
 {
@@ -303,11 +298,10 @@ fu_steelseries_sonic_read_from_flash(FuDevice *device,
 					       FALSE,
 					       error))
 			return FALSE;
-		if (g_getenv("FWUPD_STEELSERIES_SONIC_VERBOSE") != NULL)
-			fu_dump_raw(G_LOG_DOMAIN, "ReadFromFlash", data, sizeof(data));
+		fu_dump_raw(G_LOG_DOMAIN, "ReadFromFlash", data, sizeof(data));
 
 		/* timeout to give some time to read from flash to ram */
-		g_usleep(15000); /* 15 ms */
+		fu_device_sleep(device, 15); /* ms */
 
 		if (!fu_steelseries_sonic_read_from_ram(device,
 							chip,
@@ -317,7 +311,6 @@ fu_steelseries_sonic_read_from_flash(FuDevice *device,
 							fu_progress_get_child(progress),
 							error))
 			return FALSE;
-
 		fu_progress_step_done(progress);
 	}
 
@@ -329,23 +322,25 @@ static gboolean
 fu_steelseries_sonic_write_to_ram(FuDevice *device,
 				  SteelseriesSonicChip chip,
 				  guint16 address,
-				  const guint8 *buf,
-				  guint16 bufsz,
+				  GBytes *fw,
 				  FuProgress *progress,
 				  GError **error)
 {
 	const guint16 opcode = STEELSERIES_SONIC_WRITE_TO_RAM_OPCODE[chip];
 	guint8 data[STEELSERIES_BUFFER_CONTROL_SIZE] = {0};
-	g_autoptr(GPtrArray) chunks = NULL;
+	g_autoptr(FuChunkArray) chunks = NULL;
 
-	chunks = fu_chunk_array_new(buf, bufsz, 0x0, 0x0, STEELSERIES_BUFFER_RAM_TRANSFER_SIZE);
+	chunks = fu_chunk_array_new_from_bytes(fw, 0x0, STEELSERIES_BUFFER_RAM_TRANSFER_SIZE);
 	fu_progress_set_id(progress, G_STRLOC);
 	fu_progress_set_status(progress, FWUPD_STATUS_DEVICE_WRITE);
-	fu_progress_set_steps(progress, chunks->len);
-	for (guint i = 0; i < chunks->len; i++) {
-		FuChunk *chk = g_ptr_array_index(chunks, i);
-		const guint16 offset = fu_chunk_get_address(chk);
-		const guint16 size = fu_chunk_get_data_sz(chk);
+	fu_progress_set_steps(progress, fu_chunk_array_length(chunks));
+	for (guint i = 0; i < fu_chunk_array_length(chunks); i++) {
+		g_autoptr(FuChunk) chk = NULL;
+
+		/* prepare chunk */
+		chk = fu_chunk_array_index(chunks, i, error);
+		if (chk == NULL)
+			return FALSE;
 
 		if (!fu_memwrite_uint16_safe(data,
 					     sizeof(data),
@@ -358,7 +353,7 @@ fu_steelseries_sonic_write_to_ram(FuDevice *device,
 		if (!fu_memwrite_uint16_safe(data,
 					     sizeof(data),
 					     STEELSERIES_SONIC_WRITE_TO_RAM_OFFSET_OFFSET,
-					     offset,
+					     fu_chunk_get_address(chk),
 					     G_LITTLE_ENDIAN,
 					     error))
 			return FALSE;
@@ -366,7 +361,7 @@ fu_steelseries_sonic_write_to_ram(FuDevice *device,
 		if (!fu_memwrite_uint16_safe(data,
 					     sizeof(data),
 					     STEELSERIES_SONIC_WRITE_TO_RAM_SIZE_OFFSET,
-					     size,
+					     fu_chunk_get_data_sz(chk),
 					     G_LITTLE_ENDIAN,
 					     error))
 			return FALSE;
@@ -381,8 +376,7 @@ fu_steelseries_sonic_write_to_ram(FuDevice *device,
 				    error))
 			return FALSE;
 
-		if (g_getenv("FWUPD_STEELSERIES_SONIC_VERBOSE") != NULL)
-			fu_dump_raw(G_LOG_DOMAIN, "WriteToRAM", data, sizeof(data));
+		fu_dump_raw(G_LOG_DOMAIN, "WriteToRAM", data, sizeof(data));
 		if (!fu_steelseries_device_cmd(FU_STEELSERIES_DEVICE(device),
 					       data,
 					       sizeof(data),
@@ -391,8 +385,7 @@ fu_steelseries_sonic_write_to_ram(FuDevice *device,
 			return FALSE;
 
 		/* timeout to give some time to write to ram */
-		g_usleep(15000); /* 15 ms */
-
+		fu_device_sleep(device, 15); /* ms */
 		fu_progress_step_done(progress);
 	}
 
@@ -404,30 +397,33 @@ static gboolean
 fu_steelseries_sonic_write_to_flash(FuDevice *device,
 				    SteelseriesSonicChip chip,
 				    guint32 address,
-				    const guint8 *buf,
-				    guint32 bufsz,
+				    GBytes *fw,
 				    FuProgress *progress,
 				    GError **error)
 {
 	const guint16 opcode = STEELSERIES_SONIC_WRITE_TO_FLASH_OPCODE[chip];
 	const guint16 chipid = STEELSERIES_SONIC_CHIP[chip];
 	guint8 data[STEELSERIES_BUFFER_CONTROL_SIZE] = {0};
-	g_autoptr(GPtrArray) chunks = NULL;
+	g_autoptr(FuChunkArray) chunks = NULL;
 
-	chunks = fu_chunk_array_new(buf, bufsz, 0x0, 0x0, STEELSERIES_BUFFER_FLASH_TRANSFER_SIZE);
+	chunks = fu_chunk_array_new_from_bytes(fw, 0x0, STEELSERIES_BUFFER_FLASH_TRANSFER_SIZE);
 	fu_progress_set_id(progress, G_STRLOC);
 	fu_progress_set_status(progress, FWUPD_STATUS_DEVICE_WRITE);
-	fu_progress_set_steps(progress, chunks->len);
-	for (guint i = 0; i < chunks->len; i++) {
-		FuChunk *chk = g_ptr_array_index(chunks, i);
-		const guint32 offset = fu_chunk_get_address(chk);
-		const guint16 size = fu_chunk_get_data_sz(chk);
+	fu_progress_set_steps(progress, fu_chunk_array_length(chunks));
+	for (guint i = 0; i < fu_chunk_array_length(chunks); i++) {
+		g_autoptr(FuChunk) chk = NULL;
+		g_autoptr(GBytes) chk_blob = NULL;
 
+		/* prepare chunk */
+		chk = fu_chunk_array_index(chunks, i, error);
+		if (chk == NULL)
+			return FALSE;
+
+		chk_blob = fu_chunk_get_bytes(chk);
 		if (!fu_steelseries_sonic_write_to_ram(device,
 						       chip,
-						       offset,
-						       fu_chunk_get_data(chk),
-						       size,
+						       fu_chunk_get_address(chk),
+						       chk_blob,
 						       fu_progress_get_child(progress),
 						       error))
 			return FALSE;
@@ -451,7 +447,7 @@ fu_steelseries_sonic_write_to_flash(FuDevice *device,
 		if (!fu_memwrite_uint32_safe(data,
 					     sizeof(data),
 					     STEELSERIES_SONIC_WRITE_TO_FLASH_OFFSET_OFFSET,
-					     offset,
+					     fu_chunk_get_address(chk),
 					     G_LITTLE_ENDIAN,
 					     error))
 			return FALSE;
@@ -459,13 +455,12 @@ fu_steelseries_sonic_write_to_flash(FuDevice *device,
 		if (!fu_memwrite_uint16_safe(data,
 					     sizeof(data),
 					     STEELSERIES_SONIC_WRITE_TO_FLASH_SIZE_OFFSET,
-					     size,
+					     fu_chunk_get_data_sz(chk),
 					     G_LITTLE_ENDIAN,
 					     error))
 			return FALSE;
 
-		if (g_getenv("FWUPD_STEELSERIES_SONIC_VERBOSE") != NULL)
-			fu_dump_raw(G_LOG_DOMAIN, "WriteToFlash", data, sizeof(data));
+		fu_dump_raw(G_LOG_DOMAIN, "WriteToFlash", data, sizeof(data));
 		if (!fu_steelseries_device_cmd(FU_STEELSERIES_DEVICE(device),
 					       data,
 					       sizeof(data),
@@ -474,8 +469,7 @@ fu_steelseries_sonic_write_to_flash(FuDevice *device,
 			return FALSE;
 
 		/* timeout to give some time to write from ram to flash */
-		g_usleep(15000); /* 15 ms */
-
+		fu_device_sleep(device, 15); /* ms */
 		fu_progress_step_done(progress);
 	}
 
@@ -513,8 +507,7 @@ fu_steelseries_sonic_erase(FuDevice *device,
 				     error))
 		return FALSE;
 
-	if (g_getenv("FWUPD_STEELSERIES_SONIC_VERBOSE") != NULL)
-		fu_dump_raw(G_LOG_DOMAIN, "Erase", data, sizeof(data));
+	fu_dump_raw(G_LOG_DOMAIN, "Erase", data, sizeof(data));
 	if (!fu_steelseries_device_cmd(FU_STEELSERIES_DEVICE(device),
 				       data,
 				       sizeof(data),
@@ -523,8 +516,7 @@ fu_steelseries_sonic_erase(FuDevice *device,
 		return FALSE;
 
 	/* timeout to give some time to erase flash */
-	fu_progress_sleep(fu_progress_get_child(progress), 1000); /* 1 s */
-
+	fu_device_sleep_full(device, 1000, fu_progress_get_child(progress)); /* ms */
 	fu_progress_step_done(progress);
 
 	/* success */
@@ -552,8 +544,7 @@ fu_steelseries_sonic_restart(FuDevice *device,
 				     error))
 		return FALSE;
 
-	if (g_getenv("FWUPD_STEELSERIES_SONIC_VERBOSE") != NULL)
-		fu_dump_raw(G_LOG_DOMAIN, "Restart", data, sizeof(data));
+	fu_dump_raw(G_LOG_DOMAIN, "Restart", data, sizeof(data));
 	if (!fu_steelseries_device_cmd(FU_STEELSERIES_DEVICE(device),
 				       data,
 				       sizeof(data),
@@ -562,8 +553,7 @@ fu_steelseries_sonic_restart(FuDevice *device,
 		return FALSE;
 
 	/* timeout to give some time to restart chip */
-	fu_progress_sleep(progress, 3000); /* 3 s */
-
+	fu_device_sleep_full(device, 3000, progress); /* ms */
 	fu_progress_step_done(progress);
 
 	/* success */
@@ -590,7 +580,10 @@ fu_steelseries_sonic_wait_for_connect_cb(FuDevice *device, gpointer user_data, G
 }
 
 static gboolean
-fu_steelseries_sonic_wait_for_connect(FuDevice *device, guint delay, GError **error)
+fu_steelseries_sonic_wait_for_connect(FuDevice *device,
+				      guint delay,
+				      FuProgress *progress,
+				      GError **error)
 {
 	SteelseriesSonicWirelessStatus wl_status;
 	g_autoptr(FwupdRequest) request = NULL;
@@ -616,7 +609,8 @@ fu_steelseries_sonic_wait_for_connect(FuDevice *device, guint delay, GError **er
 	fwupd_request_set_kind(request, FWUPD_REQUEST_KIND_IMMEDIATE);
 	fwupd_request_set_id(request, FWUPD_REQUEST_ID_PRESS_UNLOCK);
 	fwupd_request_set_message(request, msg);
-	fu_device_emit_request(device, request);
+	if (!fu_device_emit_request(device, request, progress, error))
+		return FALSE;
 
 	if (!fu_device_retry_full(device,
 				  fu_steelseries_sonic_wait_for_connect_cb,
@@ -670,11 +664,11 @@ fu_steelseries_sonic_attach(FuDevice *device, FuProgress *progress, GError **err
 	fwupd_request_set_id(request, FWUPD_REQUEST_ID_REMOVE_REPLUG);
 	fwupd_request_add_flag(request, FWUPD_REQUEST_FLAG_ALLOW_GENERIC_MESSAGE);
 	fwupd_request_set_message(request, msg);
-	fu_device_emit_request(device, request);
-
-	fu_device_add_flag(device, FWUPD_DEVICE_FLAG_WAIT_FOR_REPLUG);
+	if (!fu_device_emit_request(device, request, progress, error))
+		return FALSE;
 
 	/* success */
+	fu_device_add_flag(device, FWUPD_DEVICE_FLAG_WAIT_FOR_REPLUG);
 	return TRUE;
 }
 
@@ -688,6 +682,7 @@ fu_steelseries_sonic_prepare(FuDevice *device,
 
 	if (!fu_steelseries_sonic_wait_for_connect(device,
 						   fu_device_get_remove_delay(device),
+						   progress,
 						   error))
 		return FALSE;
 
@@ -734,8 +729,7 @@ fu_steelseries_sonic_write_chip(FuDevice *device,
 	buf = fu_bytes_get_data_safe(blob, &bufsz, error);
 	if (buf == NULL)
 		return FALSE;
-	if (g_getenv("FWUPD_STEELSERIES_SONIC_VERBOSE") != NULL)
-		fu_dump_raw(G_LOG_DOMAIN, STEELSERIES_SONIC_FIRMWARE_ID[chip], buf, bufsz);
+	fu_dump_raw(G_LOG_DOMAIN, STEELSERIES_SONIC_FIRMWARE_ID[chip], buf, bufsz);
 	if (!fu_steelseries_sonic_erase(device, chip, fu_progress_get_child(progress), error)) {
 		g_prefix_error(error, "failed to erase chip %u: ", chip);
 		return FALSE;
@@ -744,8 +738,7 @@ fu_steelseries_sonic_write_chip(FuDevice *device,
 	if (!fu_steelseries_sonic_write_to_flash(device,
 						 chip,
 						 0x0,
-						 buf,
-						 bufsz,
+						 blob,
 						 fu_progress_get_child(progress),
 						 error)) {
 		g_prefix_error(error, "failed to write to flash chip %u: ", chip);
@@ -763,7 +756,7 @@ fu_steelseries_sonic_read_chip(FuDevice *device,
 			       FuProgress *progress,
 			       GError **error)
 {
-	guint32 bufsz;
+	gsize bufsz;
 	g_autoptr(GBytes) blob = NULL;
 	g_autofree guint8 *buf = NULL;
 
@@ -819,12 +812,10 @@ fu_steelseries_sonic_verify_chip(FuDevice *device,
 	if (blob_tmp == NULL)
 		return FALSE;
 	if (!fu_bytes_compare(blob_tmp, blob, error)) {
-		if (g_getenv("FWUPD_STEELSERIES_SONIC_VERBOSE") != NULL) {
-			fu_dump_raw(G_LOG_DOMAIN,
-				    "Verify",
-				    g_bytes_get_data(blob_tmp, NULL),
-				    g_bytes_get_size(blob_tmp));
-		}
+		fu_dump_raw(G_LOG_DOMAIN,
+			    "Verify",
+			    g_bytes_get_data(blob_tmp, NULL),
+			    g_bytes_get_size(blob_tmp));
 		return FALSE;
 	}
 	fu_progress_step_done(progress);
@@ -844,6 +835,7 @@ fu_steelseries_sonic_read_firmware(FuDevice *device, FuProgress *progress, GErro
 
 	if (!fu_steelseries_sonic_wait_for_connect(device,
 						   fu_device_get_remove_delay(device),
+						   progress,
 						   error))
 		return NULL;
 
@@ -984,8 +976,9 @@ fu_steelseries_sonic_parse_firmware(FuFirmware *firmware, FwupdInstallFlags flag
 				    G_LITTLE_ENDIAN,
 				    error))
 		return FALSE;
-	checksum_tmp =
-	    fu_crc32(g_bytes_get_data(blob, NULL), g_bytes_get_size(blob) - sizeof(checksum_tmp));
+	checksum_tmp = fu_crc32(FU_CRC_KIND_B32_STANDARD,
+				g_bytes_get_data(blob, NULL),
+				g_bytes_get_size(blob) - sizeof(checksum_tmp));
 	checksum_tmp = ~checksum_tmp;
 	if (checksum_tmp != checksum) {
 		if ((flags & FWUPD_INSTALL_FLAG_IGNORE_CHECKSUM) == 0) {
@@ -1011,7 +1004,8 @@ fu_steelseries_sonic_parse_firmware(FuFirmware *firmware, FwupdInstallFlags flag
 
 static FuFirmware *
 fu_steelseries_sonic_prepare_firmware(FuDevice *device,
-				      GBytes *fw,
+				      GInputStream *stream,
+				      FuProgress *progress,
 				      FwupdInstallFlags flags,
 				      GError **error)
 {
@@ -1022,7 +1016,7 @@ fu_steelseries_sonic_prepare_firmware(FuDevice *device,
 	g_autoptr(FuFirmware) firmware = NULL;
 
 	firmware = fu_archive_firmware_new();
-	if (!fu_firmware_parse(firmware, fw, flags, error))
+	if (!fu_firmware_parse_stream(firmware, stream, 0x0, flags, error))
 		return NULL;
 
 	/* mouse */
@@ -1069,14 +1063,14 @@ fu_steelseries_sonic_set_progress(FuDevice *self, FuProgress *progress)
 static void
 fu_steelseries_sonic_class_init(FuSteelseriesSonicClass *klass)
 {
-	FuDeviceClass *klass_device = FU_DEVICE_CLASS(klass);
+	FuDeviceClass *device_class = FU_DEVICE_CLASS(klass);
 
-	klass_device->attach = fu_steelseries_sonic_attach;
-	klass_device->prepare = fu_steelseries_sonic_prepare;
-	klass_device->read_firmware = fu_steelseries_sonic_read_firmware;
-	klass_device->write_firmware = fu_steelseries_sonic_write_firmware;
-	klass_device->prepare_firmware = fu_steelseries_sonic_prepare_firmware;
-	klass_device->set_progress = fu_steelseries_sonic_set_progress;
+	device_class->attach = fu_steelseries_sonic_attach;
+	device_class->prepare = fu_steelseries_sonic_prepare;
+	device_class->read_firmware = fu_steelseries_sonic_read_firmware;
+	device_class->write_firmware = fu_steelseries_sonic_write_firmware;
+	device_class->prepare_firmware = fu_steelseries_sonic_prepare_firmware;
+	device_class->set_progress = fu_steelseries_sonic_set_progress;
 }
 
 static void
@@ -1088,7 +1082,8 @@ fu_steelseries_sonic_init(FuSteelseriesSonic *self)
 	fu_device_add_flag(FU_DEVICE(self), FWUPD_DEVICE_FLAG_UPDATABLE);
 	fu_device_add_flag(FU_DEVICE(self), FWUPD_DEVICE_FLAG_USABLE_DURING_UPDATE);
 	fu_device_add_flag(FU_DEVICE(self), FWUPD_DEVICE_FLAG_CAN_VERIFY_IMAGE);
-	fu_device_add_internal_flag(FU_DEVICE(self), FU_DEVICE_INTERNAL_FLAG_REPLUG_MATCH_GUID);
+	fu_device_add_private_flag(FU_DEVICE(self), FU_DEVICE_PRIVATE_FLAG_REPLUG_MATCH_GUID);
+	fu_device_add_request_flag(FU_DEVICE(self), FWUPD_REQUEST_FLAG_NON_GENERIC_MESSAGE);
 	fu_device_add_protocol(FU_DEVICE(self), "com.steelseries.sonic");
 	fu_device_set_install_duration(FU_DEVICE(self), 120);				 /* 2 min */
 	fu_device_set_remove_delay(FU_DEVICE(self), FU_DEVICE_REMOVE_DELAY_USER_REPLUG); /* 40 s */
