@@ -1,19 +1,18 @@
 /*
- * Copyright (C) 2012 Andrew Duggan
- * Copyright (C) 2012 Synaptics Inc.
- * Copyright (C) 2019 Richard Hughes <richard@hughsie.com>
+ * Copyright 2012 Andrew Duggan
+ * Copyright 2012 Synaptics Inc.
+ * Copyright 2019 Richard Hughes <richard@hughsie.com>
  *
- * SPDX-License-Identifier: LGPL-2.1+
+ * SPDX-License-Identifier: LGPL-2.1-or-later
  */
 
 #include "config.h"
 
-#include <fwupdplugin.h>
-
 #include <string.h>
 
 #include "fu-synaptics-rmi-common.h"
 #include "fu-synaptics-rmi-firmware.h"
+#include "fu-synaptics-rmi-struct.h"
 
 typedef enum {
 	RMI_FIRMWARE_KIND_UNKNOWN = 0x00,
@@ -37,135 +36,59 @@ struct _FuSynapticsRmiFirmware {
 
 G_DEFINE_TYPE(FuSynapticsRmiFirmware, fu_synaptics_rmi_firmware, FU_TYPE_FIRMWARE)
 
-#define RMI_IMG_CHECKSUM_OFFSET		  0x00
-#define RMI_IMG_IO_OFFSET		  0x06
-#define RMI_IMG_BOOTLOADER_VERSION_OFFSET 0x07
-#define RMI_IMG_IMAGE_SIZE_OFFSET	  0x08
-#define RMI_IMG_CONFIG_SIZE_OFFSET	  0x0c
-#define RMI_IMG_PACKAGE_ID_OFFSET	  0x1a
-#define RMI_IMG_FW_BUILD_ID_OFFSET	  0x50
-#define RMI_IMG_SIGNATURE_SIZE_OFFSET	  0x54
-#define RMI_IMG_PRODUCT_ID_OFFSET	  0x10
-#define RMI_IMG_PRODUCT_INFO_OFFSET	  0x1e
-#define RMI_IMG_FW_OFFSET		  0x100
+#define RMI_IMG_FW_OFFSET 0x100
 
 #define RMI_IMG_V10_CNTR_ADDR_OFFSET 0x0c
 #define RMI_IMG_MAX_CONTAINERS	     1024
 
-typedef struct __attribute__((packed)) {
-	guint32 content_checksum;
-	guint16 container_id;
-	guint8 minor_version;
-	guint8 major_version;
-	guint8 reserved_08;
-	guint8 reserved_09;
-	guint8 reserved_0a;
-	guint8 reserved_0b;
-	guint32 container_option_flags;
-	guint32 content_options_length;
-	guint32 content_options_address;
-	guint32 content_length;
-	guint32 content_address;
-} RmiFirmwareContainerDescriptor;
-
-typedef enum {
-	RMI_FIRMWARE_CONTAINER_ID_TOP_LEVEL = 0,
-	RMI_FIRMWARE_CONTAINER_ID_UI,
-	RMI_FIRMWARE_CONTAINER_ID_UI_CONFIG,
-	RMI_FIRMWARE_CONTAINER_ID_BL,
-	RMI_FIRMWARE_CONTAINER_ID_BL_IMAGE,
-	RMI_FIRMWARE_CONTAINER_ID_BL_CONFIG,
-	RMI_FIRMWARE_CONTAINER_ID_BL_LOCKDOWN_INFO,
-	RMI_FIRMWARE_CONTAINER_ID_PERMANENT_CONFIG,
-	RMI_FIRMWARE_CONTAINER_ID_GUEST_CODE,
-	RMI_FIRMWARE_CONTAINER_ID_BL_PROTOCOL_DESCRIPTOR,
-	RMI_FIRMWARE_CONTAINER_ID_UI_PROTOCOL_DESCRIPTOR,
-	RMI_FIRMWARE_CONTAINER_ID_RMI_SELF_DISCOVERY,
-	RMI_FIRMWARE_CONTAINER_ID_RMI_PAGE_CONTENT,
-	RMI_FIRMWARE_CONTAINER_ID_GENERAL_INFORMATION,
-	RMI_FIRMWARE_CONTAINER_ID_DEVICE_CONFIG,
-	RMI_FIRMWARE_CONTAINER_ID_FLASH_CONFIG,
-	RMI_FIRMWARE_CONTAINER_ID_GUEST_SERIALIZATION,
-	RMI_FIRMWARE_CONTAINER_ID_GLOBAL_PARAMETERS,
-	RMI_FIRMWARE_CONTAINER_ID_CORE_CODE,
-	RMI_FIRMWARE_CONTAINER_ID_CORE_CONFIG,
-	RMI_FIRMWARE_CONTAINER_ID_DISPLAY_CONFIG,
-	RMI_FIRMWARE_CONTAINER_ID_EXTERNAL_TOUCH_AFE_CONFIG,
-	RMI_FIRMWARE_CONTAINER_ID_UTILITY,
-	RMI_FIRMWARE_CONTAINER_ID_UTILITY_PARAMETER,
-} RmiFirmwareContainerId;
-
-static const gchar *
-rmi_firmware_container_id_to_string(RmiFirmwareContainerId container_id)
-{
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_TOP_LEVEL)
-		return "top-level";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_UI)
-		return "ui";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_UI_CONFIG)
-		return "ui-config";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_BL)
-		return "bl";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_BL_IMAGE)
-		return "bl-image";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_BL_CONFIG)
-		return "bl-config";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_BL_LOCKDOWN_INFO)
-		return "bl-lockdown-info";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_PERMANENT_CONFIG)
-		return "permanent-config";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_GUEST_CODE)
-		return "guest-code";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_BL_PROTOCOL_DESCRIPTOR)
-		return "bl-protocol-descriptor";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_UI_PROTOCOL_DESCRIPTOR)
-		return "ui-protocol-descriptor";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_RMI_SELF_DISCOVERY)
-		return "rmi-self-discovery";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_RMI_PAGE_CONTENT)
-		return "rmi-page-content";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_GENERAL_INFORMATION)
-		return "general-information";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_DEVICE_CONFIG)
-		return "device-config";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_FLASH_CONFIG)
-		return "flash-config";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_GUEST_SERIALIZATION)
-		return "guest-serialization";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_GLOBAL_PARAMETERS)
-		return "global-parameters";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_CORE_CODE)
-		return "core-code";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_CORE_CONFIG)
-		return "core-config";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_DISPLAY_CONFIG)
-		return "display-config";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_EXTERNAL_TOUCH_AFE_CONFIG)
-		return "external-touch-afe-config";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_UTILITY)
-		return "utility";
-	if (container_id == RMI_FIRMWARE_CONTAINER_ID_UTILITY_PARAMETER)
-		return "utility-parameter";
-	return NULL;
-}
-
 static gboolean
 fu_synaptics_rmi_firmware_add_image(FuFirmware *firmware,
 				    const gchar *id,
-				    GBytes *fw,
+				    GInputStream *stream,
 				    gsize offset,
-				    gsize sz,
+				    gsize bufsz,
 				    GError **error)
 {
-	g_autoptr(GBytes) bytes = NULL;
-	g_autoptr(FuFirmware) img = NULL;
-
-	bytes = fu_bytes_new_offset(fw, offset, sz, error);
-	if (bytes == NULL)
+	g_autoptr(FuFirmware) img = fu_firmware_new();
+	g_autoptr(GInputStream) partial_stream = NULL;
+	partial_stream = fu_partial_input_stream_new(stream, offset, bufsz, error);
+	if (partial_stream == NULL)
+		return FALSE;
+	if (!fu_firmware_parse_stream(img, partial_stream, 0x0, FWUPD_INSTALL_FLAG_NONE, error))
 		return FALSE;
-	img = fu_firmware_new_from_bytes(bytes);
 	fu_firmware_set_id(img, id);
-	fu_firmware_add_image(firmware, img);
+	return fu_firmware_add_image_full(firmware, img, error);
+}
+
+static gboolean
+fu_synaptics_rmi_firmware_add_image_v10(FuFirmware *firmware,
+					const gchar *id,
+					GInputStream *stream,
+					gsize offset,
+					gsize bufsz,
+					gsize sig_sz,
+					GError **error)
+{
+	if (!fu_synaptics_rmi_firmware_add_image(firmware, id, stream, offset, bufsz, error))
+		return FALSE;
+	if (sig_sz != 0) {
+		g_autoptr(GInputStream) partial_stream = NULL;
+		g_autoptr(FuFirmware) img = fu_firmware_new();
+		g_autofree gchar *sig_id = NULL;
+
+		partial_stream = fu_partial_input_stream_new(stream, offset + bufsz, sig_sz, error);
+		if (partial_stream == NULL)
+			return FALSE;
+		if (!fu_firmware_parse_stream(img,
+					      partial_stream,
+					      0x0,
+					      FWUPD_INSTALL_FLAG_NONE,
+					      error))
+			return FALSE;
+		sig_id = g_strdup_printf("%s-signature", id);
+		fu_firmware_set_id(img, sig_id);
+		fu_firmware_add_image(firmware, img);
+	}
 	return TRUE;
 }
 
@@ -189,58 +112,67 @@ fu_synaptics_rmi_firmware_export(FuFirmware *firmware,
 }
 
 static gboolean
-fu_synaptics_rmi_firmware_parse_v10(FuFirmware *firmware, GBytes *fw, GError **error)
+fu_synaptics_rmi_firmware_parse_v10(FuFirmware *firmware, GInputStream *stream, GError **error)
 {
 	FuSynapticsRmiFirmware *self = FU_SYNAPTICS_RMI_FIRMWARE(firmware);
-	RmiFirmwareContainerDescriptor desc = {0x0};
 	guint16 container_id;
 	guint32 cntrs_len;
 	guint32 offset;
 	guint32 cntr_addr;
 	guint8 product_id[RMI_PRODUCT_ID_LENGTH] = {0x0};
-	gsize sz = 0;
-	const guint8 *data = g_bytes_get_data(fw, &sz);
+	gsize bufsz = 0;
+	const guint8 *buf;
+	guint32 signature_size;
+	g_autoptr(GByteArray) st_dsc = NULL;
+	g_autoptr(GBytes) fw = NULL;
 
-	if (!fu_memread_uint32_safe(data,
-				    sz,
+	/* maybe stream later */
+	fw = fu_input_stream_read_bytes(stream, 0, G_MAXSIZE, error);
+	if (fw == NULL)
+		return FALSE;
+	buf = g_bytes_get_data(fw, &bufsz);
+
+	if (!fu_memread_uint32_safe(buf,
+				    bufsz,
 				    RMI_IMG_V10_CNTR_ADDR_OFFSET,
 				    &cntr_addr,
 				    G_LITTLE_ENDIAN,
 				    error))
 		return FALSE;
-	g_debug("v10 RmiFirmwareContainerDescriptor at 0x%x", cntr_addr);
-	if (!fu_memcpy_safe((guint8 *)&desc,
-			    sizeof(desc),
-			    0x0, /* dst */
-			    data,
-			    sz,
-			    cntr_addr, /* src */
-			    sizeof(desc),
-			    error)) {
-		g_prefix_error(error, "RmiFirmwareContainerDescriptor invalid: ");
+	g_debug("v10 RmiContainerDescriptor at 0x%x", cntr_addr);
+	st_dsc = fu_struct_rmi_container_descriptor_parse_stream(stream, cntr_addr, error);
+	if (st_dsc == NULL) {
+		g_prefix_error(error, "RmiContainerDescriptor invalid: ");
 		return FALSE;
 	}
-	container_id = GUINT16_FROM_LE(desc.container_id);
-	if (container_id != RMI_FIRMWARE_CONTAINER_ID_TOP_LEVEL) {
+	if (bufsz < sizeof(guint32) + st_dsc->len) {
+		g_set_error_literal(error,
+				    FWUPD_ERROR,
+				    FWUPD_ERROR_INVALID_FILE,
+				    "stream was too small");
+		return FALSE;
+	}
+	container_id = fu_struct_rmi_container_descriptor_get_container_id(st_dsc);
+	if (container_id != FU_RMI_CONTAINER_ID_TOP_LEVEL) {
 		g_set_error(error,
 			    FWUPD_ERROR,
 			    FWUPD_ERROR_INVALID_FILE,
 			    "toplevel container_id invalid, got 0x%x expected 0x%x",
 			    (guint)container_id,
-			    (guint)RMI_FIRMWARE_CONTAINER_ID_TOP_LEVEL);
+			    (guint)FU_RMI_CONTAINER_ID_TOP_LEVEL);
 		return FALSE;
 	}
-	offset = GUINT32_FROM_LE(desc.content_address);
-	if (offset > sz - sizeof(guint32) - sizeof(desc)) {
+	offset = fu_struct_rmi_container_descriptor_get_content_address(st_dsc);
+	if (offset > bufsz - sizeof(guint32) - st_dsc->len) {
 		g_set_error(error,
 			    FWUPD_ERROR,
 			    FWUPD_ERROR_INVALID_FILE,
 			    "image offset invalid, got 0x%x, size 0x%x",
 			    (guint)offset,
-			    (guint)sz);
+			    (guint)bufsz);
 		return FALSE;
 	}
-	cntrs_len = GUINT32_FROM_LE(desc.content_length) / 4;
+	cntrs_len = fu_struct_rmi_container_descriptor_get_content_length(st_dsc) / 4;
 	if (cntrs_len > RMI_IMG_MAX_CONTAINERS) {
 		g_set_error(error,
 			    FWUPD_ERROR,
@@ -256,83 +188,115 @@ fu_synaptics_rmi_firmware_parse_v10(FuFirmware *firmware, GBytes *fw, GError **e
 		guint32 content_addr;
 		guint32 addr;
 		guint32 length;
-		if (!fu_memread_uint32_safe(data, sz, offset, &addr, G_LITTLE_ENDIAN, error))
+		g_autoptr(GByteArray) st_dsc2 = NULL;
+
+		if (!fu_memread_uint32_safe(buf, bufsz, offset, &addr, G_LITTLE_ENDIAN, error))
 			return FALSE;
-		g_debug("parsing RmiFirmwareContainerDescriptor at 0x%x", addr);
-		if (!fu_memcpy_safe((guint8 *)&desc,
-				    sizeof(desc),
-				    0x0, /* dst */
-				    data,
-				    sz,
-				    addr, /* src */
-				    sizeof(desc),
-				    error))
+		g_debug("parsing RmiContainerDescriptor at 0x%x", addr);
+
+		st_dsc2 = fu_struct_rmi_container_descriptor_parse_stream(stream, addr, error);
+		if (st_dsc2 == NULL)
 			return FALSE;
-		container_id = GUINT16_FROM_LE(desc.container_id);
-		content_addr = GUINT32_FROM_LE(desc.content_address);
-		length = GUINT32_FROM_LE(desc.content_length);
-		g_debug("RmiFirmwareContainerDescriptor 0x%02x @ 0x%x (len 0x%x)",
+		container_id = fu_struct_rmi_container_descriptor_get_container_id(st_dsc2);
+		content_addr = fu_struct_rmi_container_descriptor_get_content_address(st_dsc2);
+		length = fu_struct_rmi_container_descriptor_get_content_length(st_dsc2);
+		signature_size = fu_struct_rmi_container_descriptor_get_signature_size(st_dsc2);
+		g_debug("RmiContainerDescriptor 0x%02x @ 0x%x (len 0x%x) sig_size 0x%x",
 			container_id,
 			content_addr,
-			length);
-		if (length == 0 || length > sz) {
+			length,
+			signature_size);
+		if (length == 0 || length > bufsz) {
 			g_set_error(error,
 				    FWUPD_ERROR,
 				    FWUPD_ERROR_INVALID_FILE,
 				    "length invalid, length 0x%x, size 0x%x",
 				    (guint)length,
-				    (guint)sz);
+				    (guint)bufsz);
 			return FALSE;
 		}
-		if (content_addr > sz - length) {
+		if (content_addr > bufsz - length) {
 			g_set_error(error,
 				    FWUPD_ERROR,
 				    FWUPD_ERROR_INVALID_FILE,
 				    "address invalid, got 0x%x (length 0x%x), size 0x%x",
 				    (guint)content_addr,
 				    (guint)length,
-				    (guint)sz);
+				    (guint)bufsz);
 			return FALSE;
 		}
 		switch (container_id) {
-		case RMI_FIRMWARE_CONTAINER_ID_BL:
-			if (!fu_memread_uint8_safe(data,
-						   sz,
+		case FU_RMI_CONTAINER_ID_BL:
+			if (!fu_memread_uint8_safe(buf,
+						   bufsz,
 						   content_addr,
 						   &self->bootloader_version,
 						   error))
 				return FALSE;
 			break;
-		case RMI_FIRMWARE_CONTAINER_ID_UI:
-		case RMI_FIRMWARE_CONTAINER_ID_CORE_CODE:
-			if (!fu_synaptics_rmi_firmware_add_image(firmware,
-								 "ui",
-								 fw,
-								 content_addr,
-								 length,
-								 error))
+		case FU_RMI_CONTAINER_ID_UI:
+		case FU_RMI_CONTAINER_ID_CORE_CODE:
+			if (!fu_synaptics_rmi_firmware_add_image_v10(firmware,
+								     "ui",
+								     stream,
+								     content_addr,
+								     length,
+								     signature_size,
+								     error))
 				return FALSE;
 			break;
-		case RMI_FIRMWARE_CONTAINER_ID_FLASH_CONFIG:
-			if (!fu_synaptics_rmi_firmware_add_image(firmware,
-								 "flash-config",
-								 fw,
-								 content_addr,
-								 length,
-								 error))
+		case FU_RMI_CONTAINER_ID_FLASH_CONFIG:
+			if (!fu_synaptics_rmi_firmware_add_image_v10(firmware,
+								     "flash-config",
+								     stream,
+								     content_addr,
+								     length,
+								     signature_size,
+								     error))
 				return FALSE;
 			break;
-		case RMI_FIRMWARE_CONTAINER_ID_UI_CONFIG:
-		case RMI_FIRMWARE_CONTAINER_ID_CORE_CONFIG:
-			if (!fu_synaptics_rmi_firmware_add_image(firmware,
-								 "config",
-								 fw,
-								 content_addr,
-								 length,
-								 error))
+		case FU_RMI_CONTAINER_ID_UI_CONFIG:
+		case FU_RMI_CONTAINER_ID_CORE_CONFIG:
+			if (!fu_synaptics_rmi_firmware_add_image_v10(firmware,
+								     "config",
+								     stream,
+								     content_addr,
+								     length,
+								     signature_size,
+								     error))
 				return FALSE;
 			break;
-		case RMI_FIRMWARE_CONTAINER_ID_GENERAL_INFORMATION:
+		case FU_RMI_CONTAINER_ID_FIXED_LOCATION_DATA:
+			if (!fu_synaptics_rmi_firmware_add_image_v10(firmware,
+								     "fixed-location-data",
+								     stream,
+								     content_addr,
+								     length,
+								     signature_size,
+								     error))
+				return FALSE;
+			break;
+		case FU_RMI_CONTAINER_ID_EXTERNAL_TOUCH_AFE_CONFIG:
+			if (!fu_synaptics_rmi_firmware_add_image_v10(firmware,
+								     "afe-config",
+								     stream,
+								     content_addr,
+								     length,
+								     signature_size,
+								     error))
+				return FALSE;
+			break;
+		case FU_RMI_CONTAINER_ID_DISPLAY_CONFIG:
+			if (!fu_synaptics_rmi_firmware_add_image_v10(firmware,
+								     "display-config",
+								     stream,
+								     content_addr,
+								     length,
+								     signature_size,
+								     error))
+				return FALSE;
+			break;
+		case FU_RMI_CONTAINER_ID_GENERAL_INFORMATION:
 			if (length < 0x18 + RMI_PRODUCT_ID_LENGTH) {
 				g_set_error(error,
 					    FWUPD_ERROR,
@@ -344,15 +308,15 @@ fu_synaptics_rmi_firmware_parse_v10(FuFirmware *firmware, GBytes *fw, GError **e
 			}
 			g_clear_pointer(&self->product_id, g_free);
 			self->io = 1;
-			if (!fu_memread_uint32_safe(data,
-						    sz,
+			if (!fu_memread_uint32_safe(buf,
+						    bufsz,
 						    content_addr,
 						    &self->package_id,
 						    G_LITTLE_ENDIAN,
 						    error))
 				return FALSE;
-			if (!fu_memread_uint32_safe(data,
-						    sz,
+			if (!fu_memread_uint32_safe(buf,
+						    bufsz,
 						    content_addr + 0x04,
 						    &self->build_id,
 						    G_LITTLE_ENDIAN,
@@ -361,8 +325,8 @@ fu_synaptics_rmi_firmware_parse_v10(FuFirmware *firmware, GBytes *fw, GError **e
 			if (!fu_memcpy_safe(product_id,
 					    sizeof(product_id),
 					    0x0, /* dst */
-					    data,
-					    sz,
+					    buf,
+					    bufsz,
 					    content_addr + 0x18, /* src */
 					    sizeof(product_id),
 					    error))
@@ -370,7 +334,7 @@ fu_synaptics_rmi_firmware_parse_v10(FuFirmware *firmware, GBytes *fw, GError **e
 			break;
 		default:
 			g_debug("unsupported container %s [0x%02x]",
-				rmi_firmware_container_id_to_string(container_id),
+				fu_rmi_container_id_to_string(container_id),
 				container_id);
 			break;
 		}
@@ -384,30 +348,25 @@ fu_synaptics_rmi_firmware_parse_v10(FuFirmware *firmware, GBytes *fw, GError **e
 }
 
 static gboolean
-fu_synaptics_rmi_firmware_parse_v0x(FuFirmware *firmware, GBytes *fw, GError **error)
+fu_synaptics_rmi_firmware_parse_v0x(FuFirmware *firmware, GInputStream *stream, GError **error)
 {
 	FuSynapticsRmiFirmware *self = FU_SYNAPTICS_RMI_FIRMWARE(firmware);
 	guint32 cfg_sz;
-	guint32 img_sz = 0;
-	guint32 sig_offset = 0;
-	gsize sz = 0;
-	const guint8 *data = g_bytes_get_data(fw, &sz);
+	guint32 img_sz;
+	g_autoptr(GByteArray) st_img = NULL;
 
 	/* main firmware */
-	if (!fu_memread_uint32_safe(data,
-				    sz,
-				    RMI_IMG_IMAGE_SIZE_OFFSET,
-				    &img_sz,
-				    G_LITTLE_ENDIAN,
-				    error))
+	st_img = fu_struct_rmi_img_parse_stream(stream, 0x0, error);
+	if (st_img == NULL)
 		return FALSE;
+	img_sz = fu_struct_rmi_img_get_image_size(st_img);
 	if (img_sz > 0) {
 		/* payload, then signature appended */
 		if (self->sig_size > 0) {
-			sig_offset = img_sz - self->sig_size;
+			guint32 sig_offset = img_sz - self->sig_size;
 			if (!fu_synaptics_rmi_firmware_add_image(firmware,
 								 "sig",
-								 fw,
+								 stream,
 								 RMI_IMG_FW_OFFSET + sig_offset,
 								 self->sig_size,
 								 error))
@@ -415,7 +374,7 @@ fu_synaptics_rmi_firmware_parse_v0x(FuFirmware *firmware, GBytes *fw, GError **e
 		}
 		if (!fu_synaptics_rmi_firmware_add_image(firmware,
 							 "ui",
-							 fw,
+							 stream,
 							 RMI_IMG_FW_OFFSET,
 							 img_sz,
 							 error))
@@ -423,17 +382,11 @@ fu_synaptics_rmi_firmware_parse_v0x(FuFirmware *firmware, GBytes *fw, GError **e
 	}
 
 	/* config */
-	if (!fu_memread_uint32_safe(data,
-				    sz,
-				    RMI_IMG_CONFIG_SIZE_OFFSET,
-				    &cfg_sz,
-				    G_LITTLE_ENDIAN,
-				    error))
-		return FALSE;
+	cfg_sz = fu_struct_rmi_img_get_config_size(st_img);
 	if (cfg_sz > 0) {
 		if (!fu_synaptics_rmi_firmware_add_image(firmware,
 							 "config",
-							 fw,
+							 stream,
 							 RMI_IMG_FW_OFFSET + img_sz,
 							 cfg_sz,
 							 error))
@@ -444,43 +397,46 @@ fu_synaptics_rmi_firmware_parse_v0x(FuFirmware *firmware, GBytes *fw, GError **e
 
 static gboolean
 fu_synaptics_rmi_firmware_parse(FuFirmware *firmware,
-				GBytes *fw,
-				gsize offset,
+				GInputStream *stream,
 				FwupdInstallFlags flags,
 				GError **error)
 {
 	FuSynapticsRmiFirmware *self = FU_SYNAPTICS_RMI_FIRMWARE(firmware);
-	gsize sz = 0;
-	guint32 checksum_calculated;
-	guint32 firmware_size = 0;
-	const guint8 *data = g_bytes_get_data(fw, &sz);
+	gsize bufsz = 0;
+	const guint8 *buf;
+	g_autoptr(GByteArray) st_img = NULL;
+	g_autoptr(GBytes) fw = NULL;
 
-	/* check minimum size */
-	if (sz < RMI_IMG_FW_OFFSET) {
-		g_set_error_literal(error,
-				    FWUPD_ERROR,
-				    FWUPD_ERROR_INVALID_FILE,
-				    "not enough data to parse header");
+	/* maybe stream later */
+	fw = fu_input_stream_read_bytes(stream, 0x0, G_MAXSIZE, error);
+	if (fw == NULL)
 		return FALSE;
-	}
-	if (sz % 2 != 0) {
+	buf = g_bytes_get_data(fw, &bufsz);
+
+	/* sanity check */
+	st_img = fu_struct_rmi_img_parse_stream(stream, 0x0, error);
+	if (st_img == NULL)
+		return FALSE;
+	if (bufsz % 2 != 0) {
 		g_set_error_literal(error,
 				    FWUPD_ERROR,
 				    FWUPD_ERROR_INVALID_FILE,
 				    "data not aligned to 16 bits");
 		return FALSE;
 	}
+	if (bufsz < 4) {
+		g_set_error_literal(error,
+				    FWUPD_ERROR,
+				    FWUPD_ERROR_INVALID_FILE,
+				    "stream was too small");
+		return FALSE;
+	}
 
 	/* verify checksum */
-	if (!fu_memread_uint32_safe(data,
-				    sz,
-				    RMI_IMG_CHECKSUM_OFFSET,
-				    &self->checksum,
-				    G_LITTLE_ENDIAN,
-				    error))
-		return FALSE;
-	checksum_calculated = fu_synaptics_rmi_generate_checksum(data + 4, sz - 4);
+	self->checksum = fu_struct_rmi_img_get_checksum(st_img);
 	if ((flags & FWUPD_INSTALL_FLAG_IGNORE_CHECKSUM) == 0) {
+		guint32 checksum_calculated =
+		    fu_synaptics_rmi_generate_checksum(buf + 4, bufsz - 4);
 		if (self->checksum != checksum_calculated) {
 			g_set_error(error,
 				    FWUPD_ERROR,
@@ -494,41 +450,15 @@ fu_synaptics_rmi_firmware_parse(FuFirmware *firmware,
 
 	/* parse legacy image */
 	g_clear_pointer(&self->product_id, g_free);
-	self->io = data[RMI_IMG_IO_OFFSET];
-	self->bootloader_version = data[RMI_IMG_BOOTLOADER_VERSION_OFFSET];
+	self->io = fu_struct_rmi_img_get_io_offset(st_img);
+	self->bootloader_version = fu_struct_rmi_img_get_bootloader_version(st_img);
 	if (self->io == 1) {
-		if (!fu_memread_uint32_safe(data,
-					    sz,
-					    RMI_IMG_FW_BUILD_ID_OFFSET,
-					    &self->build_id,
-					    G_LITTLE_ENDIAN,
-					    error))
-			return FALSE;
-		if (!fu_memread_uint32_safe(data,
-					    sz,
-					    RMI_IMG_PACKAGE_ID_OFFSET,
-					    &self->package_id,
-					    G_LITTLE_ENDIAN,
-					    error))
-			return FALSE;
+		self->build_id = fu_struct_rmi_img_get_fw_build_id(st_img);
+		self->package_id = fu_struct_rmi_img_get_package_id(st_img);
 	}
-	self->product_id =
-	    g_strndup((const gchar *)data + RMI_IMG_PRODUCT_ID_OFFSET, RMI_PRODUCT_ID_LENGTH);
-	if (!fu_memread_uint16_safe(data,
-				    sz,
-				    RMI_IMG_PRODUCT_INFO_OFFSET,
-				    &self->product_info,
-				    G_LITTLE_ENDIAN,
-				    error))
-		return FALSE;
-	if (!fu_memread_uint32_safe(data,
-				    sz,
-				    RMI_IMG_IMAGE_SIZE_OFFSET,
-				    &firmware_size,
-				    G_LITTLE_ENDIAN,
-				    error))
-		return FALSE;
-	fu_firmware_set_size(firmware, firmware_size);
+	self->product_id = fu_struct_rmi_img_get_product_id(st_img);
+	self->product_info = fu_struct_rmi_img_get_product_info(st_img);
+	fu_firmware_set_size(firmware, fu_struct_rmi_img_get_image_size(st_img));
 
 	/* parse partitions, but ignore lockdown */
 	switch (self->bootloader_version) {
@@ -537,21 +467,15 @@ fu_synaptics_rmi_firmware_parse(FuFirmware *firmware,
 	case 4:
 	case 5:
 	case 6:
-		if ((self->io & 0x10) >> 1) {
-			if (!fu_memread_uint32_safe(data,
-						    sz,
-						    RMI_IMG_SIGNATURE_SIZE_OFFSET,
-						    &self->sig_size,
-						    G_LITTLE_ENDIAN,
-						    error))
-				return FALSE;
-		}
-		if (!fu_synaptics_rmi_firmware_parse_v0x(firmware, fw, error))
+		if ((self->io & 0x10) >> 1)
+			self->sig_size = fu_struct_rmi_img_get_signature_size(st_img);
+		if (!fu_synaptics_rmi_firmware_parse_v0x(firmware, stream, error))
 			return FALSE;
 		self->kind = RMI_FIRMWARE_KIND_0X;
 		break;
 	case 16:
-		if (!fu_synaptics_rmi_firmware_parse_v10(firmware, fw, error))
+	case 17:
+		if (!fu_synaptics_rmi_firmware_parse_v10(firmware, stream, error))
 			return FALSE;
 		self->kind = RMI_FIRMWARE_KIND_10;
 		break;
@@ -574,7 +498,7 @@ fu_synaptics_rmi_firmware_get_sig_size(FuSynapticsRmiFirmware *self)
 	return self->sig_size;
 }
 
-static GBytes *
+static GByteArray *
 fu_synaptics_rmi_firmware_write_v0x(FuFirmware *firmware, GError **error)
 {
 	FuSynapticsRmiFirmware *self = FU_SYNAPTICS_RMI_FIRMWARE(firmware);
@@ -582,6 +506,7 @@ fu_synaptics_rmi_firmware_write_v0x(FuFirmware *firmware, GError **error)
 	guint32 csum;
 	g_autoptr(FuFirmware) img = NULL;
 	g_autoptr(GByteArray) buf = g_byte_array_new();
+	g_autoptr(GByteArray) st_img = fu_struct_rmi_img_new();
 	g_autoptr(GBytes) buf_blob = NULL;
 
 	/* default image */
@@ -594,38 +519,30 @@ fu_synaptics_rmi_firmware_write_v0x(FuFirmware *firmware, GError **error)
 	bufsz = g_bytes_get_size(buf_blob);
 
 	/* create empty block */
-	fu_byte_array_set_size(buf, RMI_IMG_FW_OFFSET + 0x4 + bufsz, 0x00);
-	buf->data[RMI_IMG_IO_OFFSET] = 0x0;		    /* no build_id or package_id */
-	buf->data[RMI_IMG_BOOTLOADER_VERSION_OFFSET] = 0x2; /* not hierarchical */
+	fu_struct_rmi_img_set_bootloader_version(st_img, 0x2); /* not hierarchical */
 	if (self->product_id != NULL) {
-		gsize product_id_sz = strlen(self->product_id);
-		if (!fu_memcpy_safe(buf->data,
-				    buf->len,
-				    RMI_IMG_PRODUCT_ID_OFFSET, /* dst */
-				    (const guint8 *)self->product_id,
-				    product_id_sz,
-				    0x0, /* src */
-				    product_id_sz,
-				    error))
+		if (!fu_struct_rmi_img_set_product_id(st_img, self->product_id, error))
 			return NULL;
 	}
-	fu_memwrite_uint16(buf->data + RMI_IMG_PRODUCT_INFO_OFFSET, 0x1234, G_LITTLE_ENDIAN);
-	fu_memwrite_uint32(buf->data + RMI_IMG_IMAGE_SIZE_OFFSET, bufsz, G_LITTLE_ENDIAN);
-	fu_memwrite_uint32(buf->data + RMI_IMG_CONFIG_SIZE_OFFSET, bufsz, G_LITTLE_ENDIAN);
-	fu_memwrite_uint32(buf->data + RMI_IMG_FW_OFFSET + 0x0, 0xdead, G_LITTLE_ENDIAN); /* img */
+	fu_struct_rmi_img_set_product_info(st_img, 0x1234);
+	fu_struct_rmi_img_set_image_size(st_img, bufsz);
+	fu_struct_rmi_img_set_config_size(st_img, bufsz);
+	g_byte_array_append(buf, st_img->data, st_img->len);
+	fu_byte_array_set_size(buf, RMI_IMG_FW_OFFSET + 0x4 + bufsz, 0x00);
+	fu_memwrite_uint32(buf->data + RMI_IMG_FW_OFFSET, 0xDEAD, G_LITTLE_ENDIAN); /* img */
 	fu_memwrite_uint32(buf->data + RMI_IMG_FW_OFFSET + bufsz,
-			   0xbeef,
+			   0xBEEF,
 			   G_LITTLE_ENDIAN); /* config */
 
 	/* fixup checksum */
 	csum = fu_synaptics_rmi_generate_checksum(buf->data + 4, buf->len - 4);
-	fu_memwrite_uint32(buf->data + RMI_IMG_CHECKSUM_OFFSET, csum, G_LITTLE_ENDIAN);
+	fu_memwrite_uint32(buf->data + FU_STRUCT_RMI_IMG_OFFSET_CHECKSUM, csum, G_LITTLE_ENDIAN);
 
 	/* success */
-	return g_byte_array_free_to_bytes(g_steal_pointer(&buf));
+	return g_steal_pointer(&buf);
 }
 
-static GBytes *
+static GByteArray *
 fu_synaptics_rmi_firmware_write_v10(FuFirmware *firmware, GError **error)
 {
 	FuSynapticsRmiFirmware *self = FU_SYNAPTICS_RMI_FIRMWARE(firmware);
@@ -633,22 +550,16 @@ fu_synaptics_rmi_firmware_write_v10(FuFirmware *firmware, GError **error)
 	guint32 csum;
 	g_autoptr(FuFirmware) img = NULL;
 	g_autoptr(GByteArray) buf = g_byte_array_new();
+	g_autoptr(GByteArray) desc_hdr = fu_struct_rmi_container_descriptor_new();
+	g_autoptr(GByteArray) desc = fu_struct_rmi_container_descriptor_new();
 	g_autoptr(GBytes) buf_blob = NULL;
 
 	/* header | desc_hdr | offset_table | desc | flash_config |
 	 *        \0x0       \0x20          \0x24  \0x44          |0x48 */
-	RmiFirmwareContainerDescriptor desc_hdr = {
-	    .container_id = GUINT16_TO_LE(RMI_FIRMWARE_CONTAINER_ID_TOP_LEVEL),
-	    .content_length = GUINT32_TO_LE(0x1 * 4), /* size of offset table in bytes */
-	    .content_address = GUINT32_TO_LE(RMI_IMG_FW_OFFSET + 0x20), /* offset to table */
-	};
 	guint32 offset_table[] = {
 	    GUINT32_TO_LE(RMI_IMG_FW_OFFSET + 0x24)}; /* offset to first descriptor */
-	RmiFirmwareContainerDescriptor desc = {
-	    .container_id = GUINT16_TO_LE(RMI_FIRMWARE_CONTAINER_ID_FLASH_CONFIG),
-	    .content_length = GUINT32_TO_LE(0x0),
-	    .content_address = GUINT32_TO_LE(RMI_IMG_FW_OFFSET + 0x44),
-	};
+	fu_struct_rmi_container_descriptor_set_container_id(desc, FU_RMI_CONTAINER_ID_FLASH_CONFIG);
+	fu_struct_rmi_container_descriptor_set_content_address(desc, RMI_IMG_FW_OFFSET + 0x44);
 
 	/* default image */
 	img = fu_firmware_get_image_by_id(firmware, "ui", error);
@@ -658,17 +569,17 @@ fu_synaptics_rmi_firmware_write_v10(FuFirmware *firmware, GError **error)
 	if (buf_blob == NULL)
 		return NULL;
 	bufsz = g_bytes_get_size(buf_blob);
-	desc.content_length = GUINT32_TO_LE(bufsz);
+	fu_struct_rmi_container_descriptor_set_content_length(desc, bufsz);
 
 	/* create empty block */
 	fu_byte_array_set_size(buf, RMI_IMG_FW_OFFSET + 0x48, 0x00);
-	buf->data[RMI_IMG_IO_OFFSET] = 0x1;
-	buf->data[RMI_IMG_BOOTLOADER_VERSION_OFFSET] = 16; /* hierarchical */
+	buf->data[FU_STRUCT_RMI_IMG_OFFSET_IO_OFFSET] = 0x1;
+	buf->data[FU_STRUCT_RMI_IMG_OFFSET_BOOTLOADER_VERSION] = 16; /* hierarchical */
 	if (self->product_id != NULL) {
 		gsize product_id_sz = strlen(self->product_id);
 		if (!fu_memcpy_safe(buf->data,
 				    buf->len,
-				    RMI_IMG_PRODUCT_ID_OFFSET, /* dst */
+				    FU_STRUCT_RMI_IMG_OFFSET_PRODUCT_ID, /* dst */
 				    (const guint8 *)self->product_id,
 				    product_id_sz,
 				    0x0, /* src */
@@ -676,29 +587,49 @@ fu_synaptics_rmi_firmware_write_v10(FuFirmware *firmware, GError **error)
 				    error))
 			return NULL;
 	}
-	fu_memwrite_uint32(buf->data + RMI_IMG_FW_BUILD_ID_OFFSET, 0x1234, G_LITTLE_ENDIAN);
-	fu_memwrite_uint32(buf->data + RMI_IMG_PACKAGE_ID_OFFSET, 0x4321, G_LITTLE_ENDIAN);
-	fu_memwrite_uint16(buf->data + RMI_IMG_PRODUCT_INFO_OFFSET, 0x3456, G_LITTLE_ENDIAN);
-	fu_memwrite_uint32(buf->data + RMI_IMG_IMAGE_SIZE_OFFSET, bufsz, G_LITTLE_ENDIAN);
-	fu_memwrite_uint32(buf->data + RMI_IMG_CONFIG_SIZE_OFFSET, bufsz, G_LITTLE_ENDIAN);
+	fu_memwrite_uint32(buf->data + FU_STRUCT_RMI_IMG_OFFSET_FW_BUILD_ID,
+			   0x1234,
+			   G_LITTLE_ENDIAN);
+	fu_memwrite_uint32(buf->data + FU_STRUCT_RMI_IMG_OFFSET_PACKAGE_ID,
+			   0x4321,
+			   G_LITTLE_ENDIAN);
+	fu_memwrite_uint16(buf->data + FU_STRUCT_RMI_IMG_OFFSET_PRODUCT_INFO,
+			   0x3456,
+			   G_LITTLE_ENDIAN);
+	fu_memwrite_uint32(buf->data + FU_STRUCT_RMI_IMG_OFFSET_IMAGE_SIZE, bufsz, G_LITTLE_ENDIAN);
+	fu_memwrite_uint32(buf->data + FU_STRUCT_RMI_IMG_OFFSET_CONFIG_SIZE,
+			   bufsz,
+			   G_LITTLE_ENDIAN);
 	fu_memwrite_uint32(buf->data + RMI_IMG_V10_CNTR_ADDR_OFFSET,
 			   RMI_IMG_FW_OFFSET,
 			   G_LITTLE_ENDIAN);
 
 	/* hierarchical section */
-	memcpy(buf->data + RMI_IMG_FW_OFFSET + 0x00, &desc_hdr, sizeof(desc_hdr));
-	memcpy(buf->data + RMI_IMG_FW_OFFSET + 0x20, offset_table, sizeof(offset_table));
-	memcpy(buf->data + RMI_IMG_FW_OFFSET + 0x24, &desc, sizeof(desc));
+	fu_struct_rmi_container_descriptor_set_container_id(desc_hdr,
+							    FU_RMI_CONTAINER_ID_TOP_LEVEL);
+	fu_struct_rmi_container_descriptor_set_content_length(desc_hdr, 0x1 * 4); /* bytes */
+	fu_struct_rmi_container_descriptor_set_content_address(desc_hdr,
+							       RMI_IMG_FW_OFFSET +
+								   0x20); /* offset to table */
+	memcpy(buf->data + RMI_IMG_FW_OFFSET + 0x00,			  /* nocheck:blocked */
+	       desc_hdr->data,
+	       desc_hdr->len);
+	memcpy(buf->data + RMI_IMG_FW_OFFSET + 0x20, /* nocheck:blocked */
+	       offset_table,
+	       sizeof(offset_table));
+	memcpy(buf->data + RMI_IMG_FW_OFFSET + 0x24, /* nocheck:blocked */
+	       desc->data,
+	       desc->len);
 	fu_memwrite_uint32(buf->data + RMI_IMG_FW_OFFSET + 0x44,
 			   0xfeed,
 			   G_LITTLE_ENDIAN); /* flash_config */
 
 	/* fixup checksum */
 	csum = fu_synaptics_rmi_generate_checksum(buf->data + 4, buf->len - 4);
-	fu_memwrite_uint32(buf->data + RMI_IMG_CHECKSUM_OFFSET, csum, G_LITTLE_ENDIAN);
+	fu_memwrite_uint32(buf->data + FU_STRUCT_RMI_IMG_OFFSET_CHECKSUM, csum, G_LITTLE_ENDIAN);
 
 	/* success */
-	return g_byte_array_free_to_bytes(g_steal_pointer(&buf));
+	return g_steal_pointer(&buf);
 }
 
 static gboolean
@@ -734,7 +665,7 @@ fu_synaptics_rmi_firmware_build(FuFirmware *firmware, XbNode *n, GError **error)
 	return TRUE;
 }
 
-static GBytes *
+static GByteArray *
 fu_synaptics_rmi_firmware_write(FuFirmware *firmware, GError **error)
 {
 	FuSynapticsRmiFirmware *self = FU_SYNAPTICS_RMI_FIRMWARE(firmware);
@@ -754,6 +685,7 @@ static void
 fu_synaptics_rmi_firmware_init(FuSynapticsRmiFirmware *self)
 {
 	fu_firmware_add_flag(FU_FIRMWARE(self), FU_FIRMWARE_FLAG_HAS_CHECKSUM);
+	fu_firmware_set_images_max(FU_FIRMWARE(self), RMI_IMG_MAX_CONTAINERS);
 }
 
 static void
@@ -768,12 +700,12 @@ static void
 fu_synaptics_rmi_firmware_class_init(FuSynapticsRmiFirmwareClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS(klass);
-	FuFirmwareClass *klass_firmware = FU_FIRMWARE_CLASS(klass);
+	FuFirmwareClass *firmware_class = FU_FIRMWARE_CLASS(klass);
 	object_class->finalize = fu_synaptics_rmi_firmware_finalize;
-	klass_firmware->parse = fu_synaptics_rmi_firmware_parse;
-	klass_firmware->export = fu_synaptics_rmi_firmware_export;
-	klass_firmware->build = fu_synaptics_rmi_firmware_build;
-	klass_firmware->write = fu_synaptics_rmi_firmware_write;
+	firmware_class->parse = fu_synaptics_rmi_firmware_parse;
+	firmware_class->export = fu_synaptics_rmi_firmware_export;
+	firmware_class->build = fu_synaptics_rmi_firmware_build;
+	firmware_class->write = fu_synaptics_rmi_firmware_write;
 }
 
 FuFirmware *
