@@ -1,15 +1,14 @@
 /*
- * Copyright (C) 2012 Andrew Duggan
- * Copyright (C) 2012 Synaptics Inc.
- * Copyright (C) 2019 Richard Hughes <richard@hughsie.com>
+ * Copyright 2012 Andrew Duggan
+ * Copyright 2012 Synaptics Inc.
+ * Copyright 2019 Richard Hughes <richard@hughsie.com>
  *
- * SPDX-License-Identifier: LGPL-2.1+
+ * SPDX-License-Identifier: LGPL-2.1-or-later
  */
 
 #include "config.h"
 
-#include <fwupdplugin.h>
-
+#include "fu-synaptics-rmi-struct.h"
 #include "fu-synaptics-rmi-v7-device.h"
 
 #define RMI_F34_ERASE_WAIT_MS 10000 /* ms */
@@ -22,56 +21,9 @@ typedef enum {
 	RMI_FLASH_CMD_ERASE,
 	RMI_FLASH_CMD_ERASE_AP,
 	RMI_FLASH_CMD_SENSOR_ID,
+	RMI_FLASH_CMD_SIGNATURE,
 } RmiFlashCommand;
 
-typedef enum {
-	RMI_PARTITION_ID_NONE = 0x00,
-	RMI_PARTITION_ID_BOOTLOADER = 0x01,
-	RMI_PARTITION_ID_DEVICE_CONFIG,
-	RMI_PARTITION_ID_FLASH_CONFIG,
-	RMI_PARTITION_ID_MANUFACTURING_BLOCK,
-	RMI_PARTITION_ID_GUEST_SERIALIZATION,
-	RMI_PARTITION_ID_GLOBAL_PARAMETERS,
-	RMI_PARTITION_ID_CORE_CODE,
-	RMI_PARTITION_ID_CORE_CONFIG,
-	RMI_PARTITION_ID_GUEST_CODE,
-	RMI_PARTITION_ID_DISPLAY_CONFIG,
-	RMI_PARTITION_ID_EXTERNAL_TOUCH_AFE_CONFIG,
-	RMI_PARTITION_ID_UTILITY_PARAMETER,
-} RmiPartitionId;
-
-static const gchar *
-rmi_firmware_partition_id_to_string(RmiPartitionId partition_id)
-{
-	if (partition_id == RMI_PARTITION_ID_NONE)
-		return "none";
-	if (partition_id == RMI_PARTITION_ID_BOOTLOADER)
-		return "bootloader";
-	if (partition_id == RMI_PARTITION_ID_DEVICE_CONFIG)
-		return "device-config";
-	if (partition_id == RMI_PARTITION_ID_FLASH_CONFIG)
-		return "flash-config";
-	if (partition_id == RMI_PARTITION_ID_MANUFACTURING_BLOCK)
-		return "manufacturing-block";
-	if (partition_id == RMI_PARTITION_ID_GUEST_SERIALIZATION)
-		return "guest-serialization";
-	if (partition_id == RMI_PARTITION_ID_GLOBAL_PARAMETERS)
-		return "global-parameters";
-	if (partition_id == RMI_PARTITION_ID_CORE_CODE)
-		return "core-code";
-	if (partition_id == RMI_PARTITION_ID_CORE_CONFIG)
-		return "core-config";
-	if (partition_id == RMI_PARTITION_ID_GUEST_CODE)
-		return "guest-code";
-	if (partition_id == RMI_PARTITION_ID_DISPLAY_CONFIG)
-		return "display-config";
-	if (partition_id == RMI_PARTITION_ID_EXTERNAL_TOUCH_AFE_CONFIG)
-		return "external-touch-afe-config";
-	if (partition_id == RMI_PARTITION_ID_UTILITY_PARAMETER)
-		return "utility-parameter";
-	return NULL;
-}
-
 gboolean
 fu_synaptics_rmi_v7_device_detach(FuDevice *device, FuProgress *progress, GError **error)
 {
@@ -90,7 +42,7 @@ fu_synaptics_rmi_v7_device_detach(FuDevice *device, FuProgress *progress, GError
 		return FALSE;
 
 	/* enter BL */
-	fu_byte_array_append_uint8(enable_req, RMI_PARTITION_ID_BOOTLOADER);
+	fu_byte_array_append_uint8(enable_req, FU_RMI_PARTITION_ID_BOOTLOADER);
 	fu_byte_array_append_uint32(enable_req, 0x0, G_LITTLE_ENDIAN);
 	fu_byte_array_append_uint8(enable_req, RMI_FLASH_CMD_ENTER_BL);
 	fu_byte_array_append_uint8(enable_req, flash->bootloader_id[0]);
@@ -112,7 +64,54 @@ fu_synaptics_rmi_v7_device_detach(FuDevice *device, FuProgress *progress, GError
 		return FALSE;
 	if (!fu_synaptics_rmi_device_poll_wait(self, error))
 		return FALSE;
-	g_usleep(1000 * RMI_F34_ENABLE_WAIT_MS);
+	fu_device_sleep(device, RMI_F34_ENABLE_WAIT_MS);
+	return TRUE;
+}
+
+static gboolean
+fu_synaptics_rmi_v7_device_erase_partition(FuSynapticsRmiDevice *self,
+					   guint8 partition_id,
+					   GError **error)
+{
+	FuSynapticsRmiFunction *f34;
+	FuSynapticsRmiFlash *flash = fu_synaptics_rmi_device_get_flash(self);
+	g_autoptr(GByteArray) erase_cmd = g_byte_array_new();
+
+	/* f34 */
+	f34 = fu_synaptics_rmi_device_get_function(self, 0x34, error);
+	if (f34 == NULL)
+		return FALSE;
+	fu_byte_array_append_uint8(erase_cmd, partition_id);
+	fu_byte_array_append_uint32(erase_cmd, 0x0, G_LITTLE_ENDIAN);
+	fu_byte_array_append_uint8(erase_cmd, RMI_FLASH_CMD_ERASE);
+
+	fu_byte_array_append_uint8(erase_cmd, flash->bootloader_id[0]);
+	fu_byte_array_append_uint8(erase_cmd, flash->bootloader_id[1]);
+
+	fu_device_sleep(FU_DEVICE(self), 1000); /* ms */
+	if (!fu_synaptics_rmi_device_write(self,
+					   f34->data_base + 1,
+					   erase_cmd,
+					   FU_SYNAPTICS_RMI_DEVICE_FLAG_NONE,
+					   error)) {
+		g_prefix_error(error, "failed to unlock erasing: ");
+		return FALSE;
+	}
+	fu_device_sleep(FU_DEVICE(self), 100); /* ms */
+
+	/* wait for ATTN */
+	if (!fu_synaptics_rmi_device_wait_for_idle(self,
+						   RMI_F34_ERASE_WAIT_MS,
+						   RMI_DEVICE_WAIT_FOR_IDLE_FLAG_NONE,
+						   error)) {
+		g_prefix_error(error, "failed to wait for idle: ");
+		return FALSE;
+	}
+
+	if (!fu_synaptics_rmi_device_poll_wait(self, error)) {
+		g_prefix_error(error, "failed to get flash success: ");
+		return FALSE;
+	}
 	return TRUE;
 }
 
@@ -128,9 +127,9 @@ fu_synaptics_rmi_v7_device_erase_all(FuSynapticsRmiDevice *self, GError **error)
 	if (f34 == NULL)
 		return FALSE;
 
-	fu_byte_array_append_uint8(erase_cmd, RMI_PARTITION_ID_CORE_CODE);
+	fu_byte_array_append_uint8(erase_cmd, FU_RMI_PARTITION_ID_CORE_CODE);
 	fu_byte_array_append_uint32(erase_cmd, 0x0, G_LITTLE_ENDIAN);
-	if (flash->bootloader_id[1] == 8) {
+	if (flash->bootloader_id[1] >= 8) {
 		/* For bootloader v8 */
 		fu_byte_array_append_uint8(erase_cmd, RMI_FLASH_CMD_ERASE_AP);
 	} else {
@@ -141,8 +140,8 @@ fu_synaptics_rmi_v7_device_erase_all(FuSynapticsRmiDevice *self, GError **error)
 	fu_byte_array_append_uint8(erase_cmd, flash->bootloader_id[1]);
 	/* for BL8 device, we need hold 1 seconds after querying F34 status to
 	 * avoid not get attention by following giving erase command */
-	if (flash->bootloader_id[1] == 8)
-		g_usleep(1000 * 1000);
+	if (flash->bootloader_id[1] >= 8)
+		fu_device_sleep(FU_DEVICE(self), 1000); /* ms */
 	if (!fu_synaptics_rmi_device_write(self,
 					   f34->data_base + 1,
 					   erase_cmd,
@@ -151,8 +150,8 @@ fu_synaptics_rmi_v7_device_erase_all(FuSynapticsRmiDevice *self, GError **error)
 		g_prefix_error(error, "failed to unlock erasing: ");
 		return FALSE;
 	}
-	g_usleep(1000 * 100);
-	if (flash->bootloader_id[1] == 8) {
+	fu_device_sleep(FU_DEVICE(self), 100); /* ms */
+	if (flash->bootloader_id[1] >= 8) {
 		/* wait for ATTN */
 		if (!fu_synaptics_rmi_device_wait_for_idle(self,
 							   RMI_F34_ERASE_WAIT_MS,
@@ -171,11 +170,11 @@ fu_synaptics_rmi_v7_device_erase_all(FuSynapticsRmiDevice *self, GError **error)
 	if (flash->bootloader_id[1] == 7) {
 		g_autoptr(GByteArray) erase_config_cmd = g_byte_array_new();
 
-		fu_byte_array_append_uint8(erase_config_cmd, RMI_PARTITION_ID_CORE_CONFIG);
+		fu_byte_array_append_uint8(erase_config_cmd, FU_RMI_PARTITION_ID_CORE_CONFIG);
 		fu_byte_array_append_uint32(erase_config_cmd, 0x0, G_LITTLE_ENDIAN);
 		fu_byte_array_append_uint8(erase_config_cmd, RMI_FLASH_CMD_ERASE);
 
-		g_usleep(1000 * 100);
+		fu_device_sleep(FU_DEVICE(self), 100); /* ms */
 		if (!fu_synaptics_rmi_device_write(self,
 						   f34->data_base + 1,
 						   erase_config_cmd,
@@ -186,7 +185,7 @@ fu_synaptics_rmi_v7_device_erase_all(FuSynapticsRmiDevice *self, GError **error)
 		}
 
 		/* wait for ATTN */
-		g_usleep(1000 * 100);
+		fu_device_sleep(FU_DEVICE(self), 100); /* ms */
 		if (!fu_synaptics_rmi_device_wait_for_idle(
 			self,
 			RMI_F34_ERASE_WAIT_MS,
@@ -206,22 +205,22 @@ fu_synaptics_rmi_v7_device_erase_all(FuSynapticsRmiDevice *self, GError **error)
 static gboolean
 fu_synaptics_rmi_v7_device_write_blocks(FuSynapticsRmiDevice *self,
 					guint32 address,
-					const guint8 *data,
-					guint32 datasz,
+					GBytes *fw,
 					GError **error)
 {
 	FuSynapticsRmiFlash *flash = fu_synaptics_rmi_device_get_flash(self);
-	g_autoptr(GPtrArray) chunks = NULL;
+	g_autoptr(FuChunkArray) chunks = NULL;
 
 	/* write FW blocks */
-	chunks = fu_chunk_array_new(data,
-				    datasz,
-				    0x00, /* start addr */
-				    0x00, /* page_sz */
-				    flash->block_size);
-	for (guint i = 0; i < chunks->len; i++) {
-		FuChunk *chk = g_ptr_array_index(chunks, i);
+	chunks = fu_chunk_array_new_from_bytes(fw, 0x00, flash->block_size);
+	for (guint i = 0; i < fu_chunk_array_length(chunks); i++) {
+		g_autoptr(FuChunk) chk = NULL;
 		g_autoptr(GByteArray) req = g_byte_array_new();
+
+		/* prepare chunk */
+		chk = fu_chunk_array_index(chunks, i, error);
+		if (chk == NULL)
+			return FALSE;
 		g_byte_array_append(req, fu_chunk_get_data(chk), fu_chunk_get_data_sz(chk));
 		if (!fu_synaptics_rmi_device_write(self,
 						   address,
@@ -231,7 +230,7 @@ fu_synaptics_rmi_v7_device_write_blocks(FuSynapticsRmiDevice *self,
 			g_prefix_error(error,
 				       "failed to write block @0x%x:%x: ",
 				       address,
-				       fu_chunk_get_address(chk));
+				       (guint)fu_chunk_get_address(chk));
 			return FALSE;
 		}
 	}
@@ -249,9 +248,94 @@ fu_synaptics_rmi_v7_device_write_blocks(FuSynapticsRmiDevice *self,
 	return TRUE;
 }
 
+static gboolean
+fu_synaptics_rmi_v7_device_write_partition_signature(FuSynapticsRmiDevice *self,
+						     FuFirmware *firmware,
+						     const gchar *id,
+						     FuRmiPartitionId partition_id,
+						     GError **error)
+{
+	FuSynapticsRmiFunction *f34;
+	FuSynapticsRmiFlash *flash = fu_synaptics_rmi_device_get_flash(self);
+	g_autoptr(GByteArray) req_offset = g_byte_array_new();
+	g_autoptr(FuChunkArray) chunks = NULL;
+	g_autoptr(GBytes) bytes = NULL;
+
+	/* f34 */
+	f34 = fu_synaptics_rmi_device_get_function(self, 0x34, error);
+	if (f34 == NULL)
+		return FALSE;
+
+	/*check if signature exists */
+	bytes =
+	    fu_firmware_get_image_by_id_bytes(firmware, g_strdup_printf("%s-signature", id), NULL);
+	if (bytes == NULL) {
+		return TRUE;
+	}
+
+	/* write partition signature */
+	g_info("writing partition signature %s…", fu_rmi_partition_id_to_string(partition_id));
+
+	fu_byte_array_append_uint16(req_offset, 0x0, G_LITTLE_ENDIAN);
+	if (!fu_synaptics_rmi_device_write(self,
+					   f34->data_base + 0x2,
+					   req_offset,
+					   FU_SYNAPTICS_RMI_DEVICE_FLAG_NONE,
+					   error)) {
+		g_prefix_error(error, "failed to write offset: ");
+		return FALSE;
+	}
+
+	chunks =
+	    fu_chunk_array_new_from_bytes(bytes,
+					  0x00,
+					  (gsize)flash->payload_length * (gsize)flash->block_size);
+	for (guint i = 0; i < fu_chunk_array_length(chunks); i++) {
+		g_autoptr(FuChunk) chk = NULL;
+		g_autoptr(GByteArray) req_cmd = g_byte_array_new();
+		g_autoptr(GByteArray) req_trans_sz = g_byte_array_new();
+		g_autoptr(GBytes) chk_blob = NULL;
+
+		/* prepare chunk */
+		chk = fu_chunk_array_index(chunks, i, error);
+		if (chk == NULL)
+			return FALSE;
+
+		fu_byte_array_append_uint16(req_trans_sz,
+					    fu_chunk_get_data_sz(chk) / flash->block_size,
+					    G_LITTLE_ENDIAN);
+		if (!fu_synaptics_rmi_device_write(self,
+						   f34->data_base + 0x3,
+						   req_trans_sz,
+						   FU_SYNAPTICS_RMI_DEVICE_FLAG_NONE,
+						   error)) {
+			g_prefix_error(error, "failed to write transfer length: ");
+			return FALSE;
+		}
+		fu_byte_array_append_uint8(req_cmd, RMI_FLASH_CMD_SIGNATURE);
+		if (!fu_synaptics_rmi_device_write(self,
+						   f34->data_base + 0x4,
+						   req_cmd,
+						   FU_SYNAPTICS_RMI_DEVICE_FLAG_NONE,
+						   error)) {
+			g_prefix_error(error, "failed to write signature command: ");
+			return FALSE;
+		}
+		chk_blob = fu_chunk_get_bytes(chk);
+		if (!fu_synaptics_rmi_v7_device_write_blocks(self,
+							     f34->data_base + 0x5,
+							     chk_blob,
+							     error))
+			return FALSE;
+	}
+	return TRUE;
+}
+
 static gboolean
 fu_synaptics_rmi_v7_device_write_partition(FuSynapticsRmiDevice *self,
-					   RmiPartitionId partition_id,
+					   FuFirmware *firmware,
+					   const gchar *id,
+					   FuRmiPartitionId partition_id,
 					   GBytes *bytes,
 					   FuProgress *progress,
 					   GError **error)
@@ -260,7 +344,7 @@ fu_synaptics_rmi_v7_device_write_partition(FuSynapticsRmiDevice *self,
 	FuSynapticsRmiFlash *flash = fu_synaptics_rmi_device_get_flash(self);
 	g_autoptr(GByteArray) req_offset = g_byte_array_new();
 	g_autoptr(GByteArray) req_partition_id = g_byte_array_new();
-	g_autoptr(GPtrArray) chunks = NULL;
+	g_autoptr(FuChunkArray) chunks = NULL;
 
 	/* f34 */
 	f34 = fu_synaptics_rmi_device_get_function(self, 0x34, error);
@@ -268,7 +352,7 @@ fu_synaptics_rmi_v7_device_write_partition(FuSynapticsRmiDevice *self,
 		return FALSE;
 
 	/* write partition id */
-	g_debug("writing partition %s…", rmi_firmware_partition_id_to_string(partition_id));
+	g_info("writing partition %s…", fu_rmi_partition_id_to_string(partition_id));
 	fu_byte_array_append_uint8(req_partition_id, partition_id);
 	if (!fu_synaptics_rmi_device_write(self,
 					   f34->data_base + 0x1,
@@ -291,15 +375,21 @@ fu_synaptics_rmi_v7_device_write_partition(FuSynapticsRmiDevice *self,
 	/* write partition */
 	chunks =
 	    fu_chunk_array_new_from_bytes(bytes,
-					  0x00, /* start addr */
-					  0x00, /* page_sz */
+					  0x00,
 					  (gsize)flash->payload_length * (gsize)flash->block_size);
 	fu_progress_set_id(progress, G_STRLOC);
-	fu_progress_set_steps(progress, chunks->len);
-	for (guint i = 0; i < chunks->len; i++) {
-		FuChunk *chk = g_ptr_array_index(chunks, i);
+	fu_progress_set_steps(progress, fu_chunk_array_length(chunks) + 1);
+	for (guint i = 0; i < fu_chunk_array_length(chunks); i++) {
+		g_autoptr(FuChunk) chk = NULL;
+		g_autoptr(GBytes) chk_blob = NULL;
 		g_autoptr(GByteArray) req_trans_sz = g_byte_array_new();
 		g_autoptr(GByteArray) req_cmd = g_byte_array_new();
+
+		/* prepare chunk */
+		chk = fu_chunk_array_index(chunks, i, error);
+		if (chk == NULL)
+			return FALSE;
+
 		fu_byte_array_append_uint16(req_trans_sz,
 					    fu_chunk_get_data_sz(chk) / flash->block_size,
 					    G_LITTLE_ENDIAN);
@@ -320,14 +410,148 @@ fu_synaptics_rmi_v7_device_write_partition(FuSynapticsRmiDevice *self,
 			g_prefix_error(error, "failed to flash command: ");
 			return FALSE;
 		}
+		chk_blob = fu_chunk_get_bytes(chk);
 		if (!fu_synaptics_rmi_v7_device_write_blocks(self,
 							     f34->data_base + 0x5,
-							     fu_chunk_get_data(chk),
-							     fu_chunk_get_data_sz(chk),
+							     chk_blob,
 							     error))
 			return FALSE;
 		fu_progress_step_done(progress);
 	}
+	if (!fu_synaptics_rmi_v7_device_write_partition_signature(self,
+								  firmware,
+								  id,
+								  partition_id,
+								  error))
+		return FALSE;
+	fu_progress_step_done(progress);
+	return TRUE;
+}
+
+static GBytes *
+fu_synaptics_rmi_v7_device_get_pubkey(FuSynapticsRmiDevice *self, GError **error)
+{
+	FuSynapticsRmiFlash *flash = fu_synaptics_rmi_device_get_flash(self);
+	FuSynapticsRmiFunction *f34;
+	const gsize key_size = RMI_KEY_SIZE_2K;
+	g_autoptr(GByteArray) req_addr_zero = g_byte_array_new();
+	g_autoptr(GByteArray) req_cmd = g_byte_array_new();
+	g_autoptr(GByteArray) req_partition_id = g_byte_array_new();
+	g_autoptr(GByteArray) req_transfer_length = g_byte_array_new();
+	g_autoptr(GByteArray) res = NULL;
+	g_autoptr(GByteArray) pubkey = g_byte_array_new();
+
+	/* f34 */
+	f34 = fu_synaptics_rmi_device_get_function(self, 0x34, error);
+	if (f34 == NULL)
+		return NULL;
+
+	/* set partition id for bootloader 7 */
+	fu_byte_array_append_uint8(req_partition_id, FU_RMI_PARTITION_ID_PUBKEY);
+	if (!fu_synaptics_rmi_device_write(self,
+					   f34->data_base + 0x1,
+					   req_partition_id,
+					   FU_SYNAPTICS_RMI_DEVICE_FLAG_NONE,
+					   error)) {
+		g_prefix_error(error, "failed to write flash partition id: ");
+		return NULL;
+	}
+	fu_byte_array_append_uint16(req_addr_zero, 0x0, G_LITTLE_ENDIAN);
+	if (!fu_synaptics_rmi_device_write(self,
+					   f34->data_base + 0x2,
+					   req_addr_zero,
+					   FU_SYNAPTICS_RMI_DEVICE_FLAG_NONE,
+					   error)) {
+		g_prefix_error(error, "failed to write flash config address: ");
+		return NULL;
+	}
+
+	/* set transfer length */
+	fu_byte_array_append_uint16(req_transfer_length,
+				    key_size / flash->block_size,
+				    G_LITTLE_ENDIAN);
+	if (!fu_synaptics_rmi_device_write(self,
+					   f34->data_base + 0x3,
+					   req_transfer_length,
+					   FU_SYNAPTICS_RMI_DEVICE_FLAG_NONE,
+					   error)) {
+		g_prefix_error(error, "failed to set transfer length: ");
+		return NULL;
+	}
+
+	/* set command to read */
+	fu_byte_array_append_uint8(req_cmd, RMI_FLASH_CMD_READ);
+	if (!fu_synaptics_rmi_device_write(self,
+					   f34->data_base + 0x4,
+					   req_cmd,
+					   FU_SYNAPTICS_RMI_DEVICE_FLAG_NONE,
+					   error)) {
+		g_prefix_error(error, "failed to write command to read: ");
+		return NULL;
+	}
+	if (!fu_synaptics_rmi_device_poll_wait(self, error)) {
+		g_prefix_error(error, "failed to wait: ");
+		return NULL;
+	}
+
+	/* read back entire buffer in blocks */
+	res = fu_synaptics_rmi_device_read(self, f34->data_base + 0x5, (guint32)key_size, error);
+	if (res == NULL) {
+		g_prefix_error(error, "failed to read: ");
+		return NULL;
+	}
+
+	for (guint i = 0; i < res->len; i++)
+		fu_byte_array_append_uint8(pubkey, res->data[res->len - i - 1]);
+
+	/* success */
+	return g_bytes_new(pubkey->data, pubkey->len);
+}
+
+static gboolean
+fu_synaptics_rmi_v7_device_secure_check(FuSynapticsRmiDevice *self,
+					FuFirmware *firmware,
+					GError **error)
+{
+	FuSynapticsRmiFlash *flash = fu_synaptics_rmi_device_get_flash(self);
+	g_autoptr(GBytes) pubkey = NULL;
+	g_autoptr(GPtrArray) imgs = NULL;
+
+	if (flash->bootloader_id[1] >= 10 || flash->has_pubkey == FALSE)
+		return TRUE;
+
+	pubkey = fu_synaptics_rmi_v7_device_get_pubkey(self, error);
+	if (pubkey == NULL) {
+		g_prefix_error(error, "get pubkey failed: ");
+		return FALSE;
+	}
+
+	imgs = fu_firmware_get_images(firmware);
+	for (guint i = 0; i < imgs->len; i++) {
+		FuFirmware *img = g_ptr_array_index(imgs, i);
+		const gchar *id = fu_firmware_get_id(img);
+		g_autoptr(GBytes) byte_payload = NULL;
+		g_autoptr(GBytes) byte_signature = NULL;
+		g_autofree gchar *id_signature = NULL;
+
+		if (g_str_has_suffix(id, "-signature"))
+			continue;
+		id_signature = g_strdup_printf("%s-signature", id);
+		byte_signature = fu_firmware_get_image_by_id_bytes(firmware, id_signature, NULL);
+		if (byte_signature == NULL)
+			continue;
+		byte_payload = fu_firmware_get_bytes(img, error);
+		if (byte_payload == NULL)
+			return FALSE;
+		if (!fu_synaptics_rmi_verify_sha256_signature(byte_payload,
+							      pubkey,
+							      byte_signature,
+							      error)) {
+			g_prefix_error(error, "%s secure check failed: ", id);
+			return FALSE;
+		}
+		g_info("%s signature verified successfully", id);
+	}
 	return TRUE;
 }
 
@@ -344,15 +568,52 @@ fu_synaptics_rmi_v7_device_write_firmware(FuDevice *device,
 	g_autoptr(GBytes) bytes_bin = NULL;
 	g_autoptr(GBytes) bytes_cfg = NULL;
 	g_autoptr(GBytes) bytes_flashcfg = NULL;
+	g_autoptr(GBytes) bytes_fld = NULL;
+	g_autoptr(GBytes) bytes_afe = NULL;
+	g_autoptr(GBytes) bytes_displayconfig = NULL;
 
 	/* progress */
 	fu_progress_set_id(progress, G_STRLOC);
-	fu_progress_add_flag(progress, FU_PROGRESS_FLAG_GUESSED);
-	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_BUSY, 1, "disable-sleep");
-	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_ERASE, 10, NULL);
-	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 20, "flash-config");
-	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 20, "core-code");
-	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 20, "core-config");
+	if (flash->bootloader_id[1] > 8) {
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_BUSY, 0, "disable-sleep");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_READ, 0, "verify-signature");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 1, "fixed-location-data");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 8, "flash-config");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_ERASE, 9, NULL);
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 81, "core-code");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 1, "core-config");
+		fu_progress_add_step(progress,
+				     FWUPD_STATUS_DEVICE_WRITE,
+				     0,
+				     "external-touch-afe-config");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 0, "display-config");
+	} else if (flash->bootloader_id[1] == 8) {
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_BUSY, 0, "disable-sleep");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_READ, 0, "verify-signature");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 0, "fixed-location-data");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_ERASE, 16, NULL);
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 0, "flash-config");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 81, "core-code");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 1, "core-config");
+		fu_progress_add_step(progress,
+				     FWUPD_STATUS_DEVICE_WRITE,
+				     0,
+				     "external-touch-afe-config");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 0, "display-config");
+	} else {
+		fu_progress_add_flag(progress, FU_PROGRESS_FLAG_GUESSED);
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_BUSY, 0, "disable-sleep");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_READ, 2, "verify-signature");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 2, "fixed-location-data");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_ERASE, 3, NULL);
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 89, "core-code");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 2, "core-config");
+		fu_progress_add_step(progress,
+				     FWUPD_STATUS_DEVICE_WRITE,
+				     2,
+				     "external-touch-afe-config");
+		fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 2, "display-config");
+	}
 
 	/* we should be in bootloader mode now, but check anyway */
 	if (!fu_device_has_flag(device, FWUPD_DEVICE_FLAG_IS_BOOTLOADER)) {
@@ -375,17 +636,56 @@ fu_synaptics_rmi_v7_device_write_firmware(FuDevice *device,
 	bytes_cfg = fu_firmware_get_image_by_id_bytes(firmware, "config", error);
 	if (bytes_cfg == NULL)
 		return FALSE;
-	if (flash->bootloader_id[1] == 8) {
+	if (flash->bootloader_id[1] >= 8) {
 		bytes_flashcfg = fu_firmware_get_image_by_id_bytes(firmware, "flash-config", error);
 		if (bytes_flashcfg == NULL)
 			return FALSE;
 	}
+	bytes_fld = fu_firmware_get_image_by_id_bytes(firmware, "fixed-location-data", NULL);
+	bytes_afe = fu_firmware_get_image_by_id_bytes(firmware, "afe-config", NULL);
+	bytes_displayconfig = fu_firmware_get_image_by_id_bytes(firmware, "display-config", NULL);
 
 	/* disable powersaving */
 	if (!fu_synaptics_rmi_device_disable_sleep(self, error))
 		return FALSE;
 	fu_progress_step_done(progress);
 
+	/* verify signature */
+	if (!fu_synaptics_rmi_v7_device_secure_check(self, firmware, error))
+		return FALSE;
+	fu_progress_step_done(progress);
+
+	/* write fld before erase if exists */
+	if (bytes_fld != NULL) {
+		if (!fu_synaptics_rmi_v7_device_write_partition(
+			self,
+			firmware,
+			"fixed-location-data",
+			FU_RMI_PARTITION_ID_FIXED_LOCATION_DATA,
+			bytes_fld,
+			fu_progress_get_child(progress),
+			error))
+			return FALSE;
+	}
+	fu_progress_step_done(progress);
+
+	/* write flash config for BL > v8 */
+	if (flash->bootloader_id[1] > 8) {
+		if (!fu_synaptics_rmi_v7_device_erase_partition(self,
+								FU_RMI_PARTITION_ID_FLASH_CONFIG,
+								error))
+			return FALSE;
+		if (!fu_synaptics_rmi_v7_device_write_partition(self,
+								firmware,
+								"flash-config",
+								FU_RMI_PARTITION_ID_FLASH_CONFIG,
+								bytes_flashcfg,
+								fu_progress_get_child(progress),
+								error))
+			return FALSE;
+		fu_progress_step_done(progress);
+	}
+
 	/* erase all */
 	if (!fu_synaptics_rmi_v7_device_erase_all(self, error)) {
 		g_prefix_error(error, "failed to erase all: ");
@@ -394,19 +694,23 @@ fu_synaptics_rmi_v7_device_write_firmware(FuDevice *device,
 	fu_progress_step_done(progress);
 
 	/* write flash config for v8 */
-	if (bytes_flashcfg != NULL) {
+	if (flash->bootloader_id[1] == 8) {
 		if (!fu_synaptics_rmi_v7_device_write_partition(self,
-								RMI_PARTITION_ID_FLASH_CONFIG,
+								firmware,
+								"flash-config",
+								FU_RMI_PARTITION_ID_FLASH_CONFIG,
 								bytes_flashcfg,
 								fu_progress_get_child(progress),
 								error))
 			return FALSE;
+		fu_progress_step_done(progress);
 	}
-	fu_progress_step_done(progress);
 
 	/* write core code */
 	if (!fu_synaptics_rmi_v7_device_write_partition(self,
-							RMI_PARTITION_ID_CORE_CODE,
+							firmware,
+							"ui",
+							FU_RMI_PARTITION_ID_CORE_CODE,
 							bytes_bin,
 							fu_progress_get_child(progress),
 							error))
@@ -415,28 +719,48 @@ fu_synaptics_rmi_v7_device_write_firmware(FuDevice *device,
 
 	/* write core config */
 	if (!fu_synaptics_rmi_v7_device_write_partition(self,
-							RMI_PARTITION_ID_CORE_CONFIG,
+							firmware,
+							"config",
+							FU_RMI_PARTITION_ID_CORE_CONFIG,
 							bytes_cfg,
 							fu_progress_get_child(progress),
 							error))
 		return FALSE;
 	fu_progress_step_done(progress);
 
+	/* write afe-config if exists */
+	if (bytes_afe != NULL) {
+		if (!fu_synaptics_rmi_v7_device_write_partition(
+			self,
+			firmware,
+			"afe-config",
+			FU_RMI_PARTITION_ID_EXTERNAL_TOUCH_AFE_CONFIG,
+			bytes_afe,
+			fu_progress_get_child(progress),
+			error))
+			return FALSE;
+	}
+	fu_progress_step_done(progress);
+
+	/* write display config if exists */
+	if (bytes_displayconfig != NULL) {
+		if (!fu_synaptics_rmi_v7_device_write_partition(self,
+								firmware,
+								"display-config",
+								FU_RMI_PARTITION_ID_DISPLAY_CONFIG,
+								bytes_displayconfig,
+								fu_progress_get_child(progress),
+								error))
+			return FALSE;
+	}
+	fu_progress_step_done(progress);
+
 	/* success */
 	return TRUE;
 }
 
-typedef struct __attribute__((packed)) {
-	guint16 partition_id;
-	guint16 partition_len;
-	guint16 partition_addr;
-	guint16 partition_prop;
-} RmiPartitionTbl;
-
-G_STATIC_ASSERT(sizeof(RmiPartitionTbl) == 8);
-
 static gboolean
-fu_synaptics_rmi_device_read_flash_config_v7(FuSynapticsRmiDevice *self, GError **error)
+fu_synaptics_rmi_v7_device_read_flash_config(FuSynapticsRmiDevice *self, GError **error)
 {
 	FuSynapticsRmiFlash *flash = fu_synaptics_rmi_device_get_flash(self);
 	FuSynapticsRmiFunction *f34;
@@ -445,6 +769,7 @@ fu_synaptics_rmi_device_read_flash_config_v7(FuSynapticsRmiDevice *self, GError
 	g_autoptr(GByteArray) req_partition_id = g_byte_array_new();
 	g_autoptr(GByteArray) req_transfer_length = g_byte_array_new();
 	g_autoptr(GByteArray) res = NULL;
+	gsize partition_size = FU_STRUCT_RMI_PARTITION_TBL_SIZE;
 
 	/* f34 */
 	f34 = fu_synaptics_rmi_device_get_function(self, 0x34, error);
@@ -452,7 +777,7 @@ fu_synaptics_rmi_device_read_flash_config_v7(FuSynapticsRmiDevice *self, GError
 		return FALSE;
 
 	/* set partition id for bootloader 7 */
-	fu_byte_array_append_uint8(req_partition_id, RMI_PARTITION_ID_FLASH_CONFIG);
+	fu_byte_array_append_uint8(req_partition_id, FU_RMI_PARTITION_ID_FLASH_CONFIG);
 	if (!fu_synaptics_rmi_device_write(self,
 					   f34->data_base + 0x1,
 					   req_partition_id,
@@ -509,39 +834,37 @@ fu_synaptics_rmi_device_read_flash_config_v7(FuSynapticsRmiDevice *self, GError
 	}
 
 	/* debugging */
-	if (g_getenv("FWUPD_SYNAPTICS_RMI_VERBOSE") != NULL) {
-		fu_dump_full(G_LOG_DOMAIN,
-			     "FlashConfig",
-			     res->data,
-			     res->len,
-			     80,
-			     FU_DUMP_FLAGS_NONE);
-	}
+	fu_dump_full(G_LOG_DOMAIN, "FlashConfig", res->data, res->len, 80, FU_DUMP_FLAGS_NONE);
+
+	if ((res->data[0] & 0x0f) == 1)
+		partition_size += 0x2;
 
 	/* parse the config length */
-	for (guint i = 0x2; i < res->len; i += sizeof(RmiPartitionTbl)) {
-		RmiPartitionTbl tbl;
-		if (!fu_memcpy_safe((guint8 *)&tbl,
-				    sizeof(tbl),
-				    0x0, /* dst */
-				    res->data,
-				    res->len,
-				    i, /* src */
-				    sizeof(tbl),
-				    error))
+	for (guint i = 0x2; i < res->len; i += partition_size) {
+		guint16 partition_id;
+		g_autoptr(GByteArray) st_prt = NULL;
+		st_prt = fu_struct_rmi_partition_tbl_parse(res->data, res->len, i, error);
+		if (st_prt == NULL)
 			return FALSE;
+		partition_id = fu_struct_rmi_partition_tbl_get_partition_id(st_prt);
 		g_debug("found partition %s (0x%02x)",
-			rmi_firmware_partition_id_to_string(tbl.partition_id),
-			tbl.partition_id);
-		if (tbl.partition_id == RMI_PARTITION_ID_CORE_CONFIG) {
-			flash->block_count_cfg = tbl.partition_len;
+			fu_rmi_partition_id_to_string(partition_id),
+			partition_id);
+		if (partition_id == FU_RMI_PARTITION_ID_CORE_CONFIG) {
+			flash->block_count_cfg =
+			    fu_struct_rmi_partition_tbl_get_partition_len(st_prt);
 			continue;
 		}
-		if (tbl.partition_id == RMI_PARTITION_ID_CORE_CODE) {
-			flash->block_count_fw = tbl.partition_len;
+		if (partition_id == FU_RMI_PARTITION_ID_CORE_CODE) {
+			flash->block_count_fw =
+			    fu_struct_rmi_partition_tbl_get_partition_len(st_prt);
 			continue;
 		}
-		if (tbl.partition_id == RMI_PARTITION_ID_NONE)
+		if (partition_id == FU_RMI_PARTITION_ID_PUBKEY) {
+			flash->has_pubkey = TRUE;
+			continue;
+		}
+		if (partition_id == FU_RMI_PARTITION_ID_NONE)
 			break;
 	}
 
@@ -625,7 +948,7 @@ fu_synaptics_rmi_v7_device_setup(FuSynapticsRmiDevice *self, GError **error)
 	}
 
 	/* read flash config */
-	return fu_synaptics_rmi_device_read_flash_config_v7(self, error);
+	return fu_synaptics_rmi_v7_device_read_flash_config(self, error);
 }
 
 gboolean
