@@ -1,7 +1,7 @@
 /*
- * Copyright (C) 2018 Richard Hughes <richard@hughsie.com>
+ * Copyright 2018 Richard Hughes <richard@hughsie.com>
  *
- * SPDX-License-Identifier: LGPL-2.1+
+ * SPDX-License-Identifier: LGPL-2.1-or-later
  */
 
 #define G_LOG_DOMAIN "FuRelease"
@@ -9,7 +9,6 @@
 #include "config.h"
 
 #include "fu-device-private.h"
-#include "fu-keyring-utils.h"
 #include "fu-release-common.h"
 #include "fu-release.h"
 
@@ -26,17 +25,50 @@ struct _FuRelease {
 	FuEngineRequest *request;
 	FuDevice *device;
 	FwupdRemote *remote;
-	FuConfig *config;
-	GBytes *blob_fw;
+	FuEngineConfig *config;
+	GInputStream *stream;
 	gchar *update_request_id;
-	FwupdReleaseFlags trust_flags;
-	gboolean is_downgrade;
+	gchar *device_version_old;
 	GPtrArray *soft_reqs; /* nullable, element-type XbNode */
 	GPtrArray *hard_reqs; /* nullable, element-type XbNode */
+	guint64 priority;
 };
 
 G_DEFINE_TYPE(FuRelease, fu_release, FWUPD_TYPE_RELEASE)
 
+static gboolean
+fu_release_ensure_trust_flags(FuRelease *self, XbNode *rel, GError **error);
+
+gchar *
+fu_release_to_string(FuRelease *self)
+{
+	const guint idt = 1;
+	g_autoptr(GString) str = g_string_new(NULL);
+
+	g_return_val_if_fail(FU_IS_RELEASE(self), NULL);
+
+	/* parent */
+	fwupd_codec_add_string(FWUPD_CODEC(self), 0, str);
+
+	/* instance */
+	if (self->request != NULL)
+		fwupd_codec_add_string(FWUPD_CODEC(self->request), idt, str);
+	if (self->device != NULL)
+		fwupd_codec_string_append(str, idt, "Device", fu_device_get_id(self->device));
+	fwupd_codec_string_append(str, idt, "DeviceVersionOld", self->device_version_old);
+	if (self->remote != NULL)
+		fwupd_codec_string_append(str, idt, "Remote", fwupd_remote_get_id(self->remote));
+	fwupd_codec_string_append_bool(str, idt, "HasConfig", self->config != NULL);
+	fwupd_codec_string_append_bool(str, idt, "HasStream", self->stream != NULL);
+	fwupd_codec_string_append(str, idt, "UpdateRequestId", self->update_request_id);
+	if (self->soft_reqs != NULL)
+		fwupd_codec_string_append_hex(str, idt, "SoftReqs", self->soft_reqs->len);
+	if (self->hard_reqs != NULL)
+		fwupd_codec_string_append_hex(str, idt, "HardReqs", self->hard_reqs->len);
+	fwupd_codec_string_append_hex(str, idt, "Priority", self->priority);
+	return g_string_free(g_steal_pointer(&str), FALSE);
+}
+
 /**
  * fu_release_set_request:
  * @self: a #FuRelease
@@ -66,6 +98,34 @@ fu_release_get_request(FuRelease *self)
 	return self->request;
 }
 
+/**
+ * fu_release_get_device_version_old:
+ * @self: a #FuRelease
+ *
+ * Gets the original [before update] device version.
+ *
+ * Returns: a string value, or %NULL if never set.
+ **/
+const gchar *
+fu_release_get_device_version_old(FuRelease *self)
+{
+	g_return_val_if_fail(FU_IS_RELEASE(self), NULL);
+	return self->device_version_old;
+}
+
+static void
+fu_release_set_device_version_old(FuRelease *self, const gchar *device_version_old)
+{
+	g_return_if_fail(FU_IS_RELEASE(self));
+
+	/* not changed */
+	if (g_strcmp0(self->device_version_old, device_version_old) == 0)
+		return;
+
+	g_free(self->device_version_old);
+	self->device_version_old = g_strdup(device_version_old);
+}
+
 /**
  * fu_release_set_device:
  * @self: a #FuRelease
@@ -78,6 +138,7 @@ fu_release_set_device(FuRelease *self, FuDevice *device)
 {
 	g_return_if_fail(FU_IS_RELEASE(self));
 	g_set_object(&self->device, device);
+	fu_release_set_device_version_old(self, fu_device_get_version(device));
 }
 
 /**
@@ -96,18 +157,18 @@ fu_release_get_device(FuRelease *self)
 }
 
 /**
- * fu_release_get_fw_blob:
+ * fu_release_get_stream:
  * @self: a #FuRelease
  *
- * Gets the firmware payload to use when installing this release.
+ * Gets the firmware stream to use when installing this release.
  *
  * Returns: (transfer none) (nullable): data
  **/
-GBytes *
-fu_release_get_fw_blob(FuRelease *self)
+GInputStream *
+fu_release_get_stream(FuRelease *self)
 {
 	g_return_val_if_fail(FU_IS_RELEASE(self), NULL);
-	return self->blob_fw;
+	return self->stream;
 }
 
 /**
@@ -187,13 +248,13 @@ fu_release_set_remote(FuRelease *self, FwupdRemote *remote)
 /**
  * fu_release_set_config:
  * @self: a #FuRelease
- * @config: (nullable): a #FuConfig
+ * @config: (nullable): a #FuEngineConfig
  *
  * Sets the config to use when loading. The config may be used for things like ordering attributes
  *like protocol priority.
  **/
 void
-fu_release_set_config(FuRelease *self, FuConfig *config)
+fu_release_set_config(FuRelease *self, FuEngineConfig *config)
 {
 	g_return_if_fail(FU_IS_RELEASE(self));
 	g_set_object(&self->config, config);
@@ -235,7 +296,12 @@ fu_release_get_release_version(FuRelease *self, const gchar *version, GError **e
 		return g_strdup(version);
 
 	/* parse as integer */
-	if (!fu_strtoull(version, &ver_uint32, 1, G_MAXUINT32, &error_local)) {
+	if (!fu_strtoull(version,
+			 &ver_uint32,
+			 1,
+			 G_MAXUINT32,
+			 FU_INTEGER_BASE_AUTO,
+			 &error_local)) {
 		g_warning("invalid release version %s: %s", version, error_local->message);
 		return g_strdup(version);
 	}
@@ -244,6 +310,79 @@ fu_release_get_release_version(FuRelease *self, const gchar *version, GError **e
 	return fu_version_from_uint32((guint32)ver_uint32, fmt);
 }
 
+static gboolean
+fu_release_load_test_result(FuRelease *self, XbNode *n, GError **error)
+{
+	const gchar *tmp;
+	g_autoptr(FwupdReport) report = fwupd_report_new();
+	g_autoptr(GPtrArray) custom = NULL;
+	g_autoptr(XbNode) os = NULL;
+	g_autoptr(XbNode) vendor_name = NULL;
+
+	tmp = xb_node_get_attr(n, "date");
+	if (tmp != NULL) {
+		g_autoptr(GDateTime) dt = NULL;
+		g_autofree gchar *iso8601 = g_strdup_printf("%sT00:00:00Z", tmp);
+		dt = g_date_time_new_from_iso8601(iso8601, NULL);
+		if (dt != NULL)
+			fwupd_report_set_created(report, g_date_time_to_unix(dt));
+	}
+	tmp = xb_node_query_text(n, "device", NULL);
+	if (tmp != NULL)
+		fwupd_report_set_device_name(report, tmp);
+	tmp = xb_node_query_text(n, "previous_version", NULL);
+	if (tmp != NULL) {
+		fwupd_report_set_version_old(report, tmp);
+		if (fu_version_compare(fu_release_get_version(self),
+				       tmp,
+				       FWUPD_VERSION_FORMAT_UNKNOWN) > 0) {
+			fwupd_report_add_flag(report, FWUPD_REPORT_FLAG_IS_UPGRADE);
+		}
+	}
+	vendor_name = xb_node_query_first(n, "vendor_name", NULL);
+	if (vendor_name != NULL) {
+		guint64 vendor_id = xb_node_get_attr_as_uint(vendor_name, "id");
+		fwupd_report_set_vendor(report, xb_node_get_text(vendor_name));
+		if (vendor_id != G_MAXUINT64)
+			fwupd_report_set_vendor_id(report, vendor_id);
+	}
+	os = xb_node_query_first(n, "os", NULL);
+	if (os != NULL) {
+		tmp = xb_node_get_attr(os, "version");
+		if (tmp != NULL)
+			fwupd_report_set_distro_version(report, tmp);
+		tmp = xb_node_get_attr(os, "variant");
+		if (tmp != NULL)
+			fwupd_report_set_distro_variant(report, tmp);
+		fwupd_report_set_distro_id(report, xb_node_get_text(os));
+	}
+	if (fu_release_get_remote_id(self) != NULL)
+		fwupd_report_set_remote_id(report, fu_release_get_remote_id(self));
+	custom = xb_node_query(n, "custom/value", 0, NULL);
+	if (custom != NULL) {
+		for (guint i = 0; i < custom->len; i++) {
+			XbNode *c = g_ptr_array_index(custom, i);
+			if (g_strcmp0(xb_node_get_attr(c, "key"), "FromOEM") == 0) {
+				fwupd_report_add_flag(report, FWUPD_REPORT_FLAG_FROM_OEM);
+				continue;
+			}
+			if (xb_node_get_attr(c, "key") == NULL || xb_node_get_text(c) == NULL) {
+				g_debug("ignoring metadata: %s=%s",
+					xb_node_get_attr(c, "key"),
+					xb_node_get_text(c));
+				continue;
+			}
+			fwupd_report_add_metadata_item(report,
+						       xb_node_get_attr(c, "key"),
+						       xb_node_get_text(c));
+		}
+	}
+
+	/* success */
+	fwupd_release_add_report(FWUPD_RELEASE(self), report);
+	return TRUE;
+}
+
 static gboolean
 fu_release_load_artifact(FuRelease *self, XbNode *artifact, GError **error)
 {
@@ -251,6 +390,7 @@ fu_release_load_artifact(FuRelease *self, XbNode *artifact, GError **error)
 	guint64 size;
 	g_autoptr(GPtrArray) locations = NULL;
 	g_autoptr(GPtrArray) checksums = NULL;
+	g_autoptr(GPtrArray) test_result = NULL;
 
 	/* filename */
 	filename = xb_node_query_text(artifact, "filename", NULL);
@@ -268,14 +408,18 @@ fu_release_load_artifact(FuRelease *self, XbNode *artifact, GError **error)
 	if (locations != NULL) {
 		for (guint i = 0; i < locations->len; i++) {
 			XbNode *n = g_ptr_array_index(locations, i);
-			g_autofree gchar *scheme = NULL;
 
 			/* check the scheme is allowed */
-			scheme = fu_release_uri_get_scheme(xb_node_get_text(n));
-			if (scheme != NULL) {
-				guint prio = fu_config_get_uri_scheme_prio(self->config, scheme);
-				if (prio == G_MAXUINT)
-					continue;
+			if (self->config != NULL) {
+				g_autofree gchar *scheme =
+				    fu_release_uri_get_scheme(xb_node_get_text(n));
+				if (scheme != NULL) {
+					guint prio =
+					    fu_engine_config_get_uri_scheme_prio(self->config,
+										 scheme);
+					if (prio == G_MAXUINT)
+						continue;
+				}
 			}
 
 			/* build the complete URI */
@@ -302,6 +446,16 @@ fu_release_load_artifact(FuRelease *self, XbNode *artifact, GError **error)
 		}
 	}
 
+	/* test results */
+	test_result = xb_node_query(artifact, "testing/test_result", 0, NULL);
+	if (test_result != NULL) {
+		for (guint i = 0; i < test_result->len; i++) {
+			XbNode *n = g_ptr_array_index(test_result, i);
+			if (!fu_release_load_test_result(self, n, error))
+				return FALSE;
+		}
+	}
+
 	/* size */
 	size = xb_node_query_text_as_uint(artifact, "size[@type='installed']", NULL);
 	if (size != G_MAXUINT64)
@@ -319,8 +473,8 @@ fu_release_scheme_compare_cb(gconstpointer a, gconstpointer b, gpointer user_dat
 	const gchar *location2 = *((const gchar **)b);
 	g_autofree gchar *scheme1 = fu_release_uri_get_scheme(location1);
 	g_autofree gchar *scheme2 = fu_release_uri_get_scheme(location2);
-	guint prio1 = fu_config_get_uri_scheme_prio(self->config, scheme1);
-	guint prio2 = fu_config_get_uri_scheme_prio(self->config, scheme2);
+	guint prio1 = fu_engine_config_get_uri_scheme_prio(self->config, scheme1);
+	guint prio2 = fu_engine_config_get_uri_scheme_prio(self->config, scheme2);
 	if (prio1 < prio2)
 		return -1;
 	if (prio1 > prio2)
@@ -519,19 +673,6 @@ fu_release_check_requirements(FuRelease *self,
 		return FALSE;
 	}
 
-	/* called with online update, test if device is supposed to allow this */
-	if ((install_flags & FWUPD_INSTALL_FLAG_OFFLINE) == 0 &&
-	    (install_flags & FWUPD_INSTALL_FLAG_FORCE) == 0 &&
-	    fu_device_has_flag(self->device, FWUPD_DEVICE_FLAG_ONLY_OFFLINE)) {
-		g_set_error(error,
-			    FWUPD_ERROR,
-			    FWUPD_ERROR_NOT_SUPPORTED,
-			    "Device %s [%s] only allows offline updates",
-			    fu_device_get_name(self->device),
-			    fu_device_get_id(self->device));
-		return FALSE;
-	}
-
 	/* get device */
 	version = fu_device_get_version(self->device);
 	if (version == NULL) {
@@ -567,7 +708,7 @@ fu_release_check_requirements(FuRelease *self,
 			    FWUPD_ERROR_NOT_SUPPORTED,
 			    "Specified firmware is older than the minimum "
 			    "required version '%s < %s'",
-			    version,
+			    fu_release_get_version(self),
 			    version_lowest);
 		return FALSE;
 	}
@@ -577,7 +718,7 @@ fu_release_check_requirements(FuRelease *self,
 				    fu_release_get_version(self),
 				    fu_device_get_version_format(self->device));
 	if (fu_device_has_flag(self->device, FWUPD_DEVICE_FLAG_ONLY_VERSION_UPGRADE) &&
-	    vercmp >= 0) {
+	    vercmp > 0) {
 		g_set_error(error,
 			    FWUPD_ERROR,
 			    FWUPD_ERROR_NOT_SUPPORTED,
@@ -592,8 +733,10 @@ fu_release_check_requirements(FuRelease *self,
 			    fu_release_get_version(self));
 		return FALSE;
 	}
-	self->is_downgrade = vercmp > 0;
-	if (self->is_downgrade && (install_flags & FWUPD_INSTALL_FLAG_ALLOW_OLDER) == 0 &&
+	if (vercmp > 0)
+		fu_release_add_flag(self, FWUPD_RELEASE_FLAG_IS_DOWNGRADE);
+	if (fu_release_has_flag(self, FWUPD_RELEASE_FLAG_IS_DOWNGRADE) &&
+	    (install_flags & FWUPD_INSTALL_FLAG_ALLOW_OLDER) == 0 &&
 	    (install_flags & FWUPD_INSTALL_FLAG_ALLOW_BRANCH_SWITCH) == 0) {
 		g_set_error(error,
 			    FWUPD_ERROR,
@@ -604,23 +747,57 @@ fu_release_check_requirements(FuRelease *self,
 		return FALSE;
 	}
 
-	/* verify */
-	if (!fu_keyring_get_release_flags(rel, &self->trust_flags, &error_local)) {
-		if (g_error_matches(error_local, FWUPD_ERROR, FWUPD_ERROR_NOT_SUPPORTED)) {
-			g_warning("Ignoring verification for %s: %s",
-				  fu_device_get_name(self->device),
-				  error_local->message);
-		} else {
-			g_propagate_error(error, g_steal_pointer(&error_local));
-			return FALSE;
-		}
-	}
+	/* success */
 	return TRUE;
 }
 
+void
+fu_release_set_priority(FuRelease *self, guint64 priority)
+{
+	g_return_if_fail(FU_IS_RELEASE(self));
+	self->priority = priority;
+}
+
+guint64
+fu_release_get_priority(FuRelease *self)
+{
+	g_return_val_if_fail(FU_IS_RELEASE(self), 0);
+	return self->priority;
+}
+
+static void
+fu_release_ensure_device_by_checksum(FuRelease *self, XbNode *component, XbNode *rel)
+{
+	g_autoptr(GPtrArray) device_checksums = NULL;
+
+	/* sanity check */
+	if (fu_device_get_checksums(self->device)->len == 0)
+		return;
+	device_checksums = xb_node_query(rel, "checksum[@target='device']", 0, NULL);
+	if (device_checksums == NULL)
+		return;
+	for (guint i = 0; i < device_checksums->len; i++) {
+		XbNode *device_checksum = g_ptr_array_index(device_checksums, i);
+		if (!fu_device_has_checksum(self->device, xb_node_get_text(device_checksum)))
+			continue;
+		fu_device_ensure_from_component(self->device, component);
+		if (fu_device_has_private_flag(self->device,
+					       FU_DEVICE_PRIVATE_FLAG_MD_SET_VERSION)) {
+			const gchar *rel_version = xb_node_get_attr(rel, "version");
+			if (rel_version == NULL)
+				continue;
+			fu_device_set_version(self->device, rel_version);
+			fu_device_remove_private_flag(self->device,
+						      FU_DEVICE_PRIVATE_FLAG_MD_SET_VERSION);
+		}
+		break;
+	}
+}
+
 /**
  * fu_release_load:
  * @self: a #FuRelease
+ * @cabinet: a #FuCabinet
  * @component: (not nullable): a #XbNode
  * @rel_optional: (nullable): a #XbNode
  * @install_flags: a #FwupdInstallFlags, e.g. %FWUPD_INSTALL_FLAG_FORCE
@@ -634,6 +811,7 @@ fu_release_check_requirements(FuRelease *self,
  **/
 gboolean
 fu_release_load(FuRelease *self,
+		FuCabinet *cabinet,
 		XbNode *component,
 		XbNode *rel_optional,
 		FwupdInstallFlags install_flags,
@@ -641,7 +819,7 @@ fu_release_load(FuRelease *self,
 {
 	const gchar *tmp;
 	guint64 tmp64;
-	GBytes *blob_fw_tmp;
+	GBytes *blob_basename;
 	g_autofree gchar *name_xpath = NULL;
 	g_autofree gchar *namevs_xpath = NULL;
 	g_autofree gchar *summary_xpath = NULL;
@@ -656,10 +834,10 @@ fu_release_load(FuRelease *self,
 	g_autoptr(GError) error_hard = NULL;
 
 	g_return_val_if_fail(FU_IS_RELEASE(self), FALSE);
+	g_return_val_if_fail(cabinet == NULL || FU_IS_CABINET(cabinet), FALSE);
 	g_return_val_if_fail(XB_IS_NODE(component), FALSE);
 	g_return_val_if_fail(rel_optional == NULL || XB_IS_NODE(rel_optional), FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
-	g_return_val_if_fail(fwupd_release_get_appstream_id(FWUPD_RELEASE(self)) == NULL, FALSE);
 
 	/* set from the component */
 	tmp = xb_node_query_text(component, "id", NULL);
@@ -693,7 +871,6 @@ fu_release_load(FuRelease *self,
 	/* use default release */
 	if (rel_optional == NULL) {
 		g_autoptr(GError) error_local = NULL;
-#if LIBXMLB_CHECK_VERSION(0, 2, 0)
 		g_autoptr(XbQuery) query = NULL;
 		query = xb_query_new_full(xb_node_get_silo(component),
 					  "releases/release",
@@ -702,9 +879,6 @@ fu_release_load(FuRelease *self,
 		if (query == NULL)
 			return FALSE;
 		rel = xb_node_query_first_full(component, query, &error_local);
-#else
-		rel = xb_node_query_first(component, "releases/release", &error_local);
-#endif
 		if (rel == NULL) {
 			g_set_error(error,
 				    FWUPD_ERROR,
@@ -717,6 +891,34 @@ fu_release_load(FuRelease *self,
 		rel = g_object_ref(rel_optional);
 	}
 
+	/* find the remote */
+	tmp = xb_node_query_text(component, "../custom/value[@key='fwupd::RemoteId']", NULL);
+	if (tmp != NULL)
+		fwupd_release_set_remote_id(FWUPD_RELEASE(self), tmp);
+	tmp = xb_node_query_text(component, "../custom/value[@key='LVFS::Distributor']", NULL);
+	if (tmp != NULL && g_str_has_prefix(tmp, "community"))
+		fwupd_release_add_flag(FWUPD_RELEASE(self), FWUPD_RELEASE_FLAG_IS_COMMUNITY);
+
+	/* use the metadata to set the device attributes */
+	if (!fu_release_ensure_trust_flags(self, rel, error))
+		return FALSE;
+	if (self->device != NULL &&
+	    fu_release_has_flag(self, FWUPD_RELEASE_FLAG_TRUSTED_METADATA)) {
+		if (fu_device_has_private_flag(self->device,
+					       FU_DEVICE_PRIVATE_FLAG_MD_ONLY_CHECKSUM)) {
+			fu_release_ensure_device_by_checksum(self, component, rel);
+		} else {
+			fu_device_ensure_from_component(self->device, component);
+		}
+	}
+
+	/* per-release priority wins, but fallback to per-component priority */
+	tmp64 = xb_node_get_attr_as_uint(rel, "priority");
+	if (tmp64 == G_MAXUINT64)
+		tmp64 = xb_node_get_attr_as_uint(component, "priority");
+	if (tmp64 != G_MAXUINT64)
+		fu_release_set_priority(self, tmp64);
+
 	/* the version is fixed up with the device format */
 	tmp = xb_node_get_attr(rel, "version");
 	if (tmp == NULL) {
@@ -736,14 +938,6 @@ fu_release_load(FuRelease *self,
 	/* optional release ID -- currently a integer but maybe namespaced in the future */
 	fwupd_release_set_id(FWUPD_RELEASE(self), xb_node_get_attr(rel, "id"));
 
-	/* find the remote */
-	tmp = xb_node_query_text(component, "../custom/value[@key='fwupd::RemoteId']", NULL);
-	if (tmp != NULL)
-		fwupd_release_set_remote_id(FWUPD_RELEASE(self), tmp);
-	tmp = xb_node_query_text(component, "../custom/value[@key='LVFS::Distributor']", NULL);
-	if (g_strcmp0(tmp, "community") == 0)
-		fwupd_release_add_flag(FWUPD_RELEASE(self), FWUPD_RELEASE_FLAG_IS_COMMUNITY);
-
 	/* this is the more modern way to do this */
 	artifact = xb_node_query_first(rel, "artifacts/artifact[@type='binary']", NULL);
 	if (artifact != NULL) {
@@ -781,7 +975,7 @@ fu_release_load(FuRelease *self,
 		if (str->len > 0)
 			fwupd_release_set_description(FWUPD_RELEASE(self), str->str);
 	}
-	if (artifact == NULL) {
+	if (fwupd_release_get_locations(FWUPD_RELEASE(self))->len == 0) {
 		tmp = xb_node_query_text(rel, "location", NULL);
 		if (tmp != NULL) {
 			g_autofree gchar *uri = NULL;
@@ -802,7 +996,7 @@ fu_release_load(FuRelease *self,
 			fwupd_release_add_location(FWUPD_RELEASE(self), uri);
 		}
 	}
-	if (artifact == NULL) {
+	if (fwupd_release_get_filename(FWUPD_RELEASE(self)) == NULL) {
 		tmp = xb_node_query_text(rel, "checksum[@target='content']", NULL);
 		if (tmp != NULL)
 			fwupd_release_set_filename(FWUPD_RELEASE(self), tmp);
@@ -813,7 +1007,7 @@ fu_release_load(FuRelease *self,
 	tmp = xb_node_query_text(rel, "url[@type='source']", NULL);
 	if (tmp != NULL)
 		fwupd_release_set_source_url(FWUPD_RELEASE(self), tmp);
-	if (artifact == NULL) {
+	if (fwupd_release_get_checksums(FWUPD_RELEASE(self))->len == 0) {
 		g_autoptr(GPtrArray) checksums = NULL;
 		checksums = xb_node_query(rel, "checksum[@target='container']", 0, NULL);
 		if (checksums != NULL) {
@@ -826,7 +1020,7 @@ fu_release_load(FuRelease *self,
 			}
 		}
 	}
-	if (artifact == NULL) {
+	if (fwupd_release_get_size(FWUPD_RELEASE(self)) == 0) {
 		tmp64 = xb_node_query_text_as_uint(rel, "size[@type='installed']", NULL);
 		if (tmp64 != G_MAXUINT64)
 			fwupd_release_set_size(FWUPD_RELEASE(self), tmp64);
@@ -924,10 +1118,20 @@ fu_release_load(FuRelease *self,
 		}
 	}
 
-	/* get per-release firmware blob */
-	blob_fw_tmp = xb_node_get_data(rel, "fwupd::FirmwareBlob");
-	if (blob_fw_tmp != NULL)
-		self->blob_fw = g_bytes_ref(blob_fw_tmp);
+	/* get per-release firmware stream */
+	blob_basename = xb_node_get_data(rel, "fwupd::FirmwareBasename");
+	if (cabinet != NULL && blob_basename != NULL) {
+		const gchar *basename = (const gchar *)g_bytes_get_data(blob_basename, NULL);
+		g_autoptr(FuFirmware) img = NULL;
+		img = fu_firmware_get_image_by_id(FU_FIRMWARE(cabinet), basename, error);
+		if (img == NULL) {
+			g_prefix_error(error, "failed to find %s: ", basename);
+			return FALSE;
+		}
+		self->stream = fu_firmware_get_stream(img, error);
+		if (self->stream == NULL)
+			return FALSE;
+	}
 
 	/* to build the firmware */
 	tmp = g_object_get_data(G_OBJECT(component), "fwupd::BuilderScript");
@@ -948,7 +1152,7 @@ fu_release_load(FuRelease *self,
 
 	/* check requirements for device */
 	if (self->device != NULL && self->request != NULL &&
-	    fu_engine_request_get_kind(self->request) == FU_ENGINE_REQUEST_KIND_ACTIVE) {
+	    !fu_engine_request_has_flag(self->request, FU_ENGINE_REQUEST_FLAG_NO_REQUIREMENTS)) {
 		if (!fu_release_check_requirements(self, component, rel, install_flags, error))
 			return FALSE;
 	}
@@ -957,22 +1161,56 @@ fu_release_load(FuRelease *self,
 	return TRUE;
 }
 
-/**
- * fu_release_get_trust_flags:
- * @self: a #FuRelease
- *
- * Gets the trust flags for this task.
- *
- * NOTE: This is only set after fu_release_load() has been called successfully, and
- * is only valid when a request has been set.
- *
- * Returns: the #FwupdReleaseFlags, e.g. #FWUPD_TRUST_FLAG_PAYLOAD
- **/
-FwupdReleaseFlags
-fu_release_get_trust_flags(FuRelease *self)
+static gboolean
+fu_release_ensure_trust_flags(FuRelease *self, XbNode *rel, GError **error)
 {
+	GBytes *blob;
+
 	g_return_val_if_fail(FU_IS_RELEASE(self), FALSE);
-	return self->trust_flags;
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	/* in the self tests */
+	if (g_getenv("FWUPD_SELF_TEST") != NULL) {
+		fu_release_add_flag(self, FWUPD_RELEASE_FLAG_TRUSTED_PAYLOAD);
+		fu_release_add_flag(self, FWUPD_RELEASE_FLAG_TRUSTED_METADATA);
+		return TRUE;
+	}
+
+	/* populated from an actual cab archive */
+	blob = g_object_get_data(G_OBJECT(rel), "fwupd::ReleaseFlags");
+	if (blob != NULL) {
+		FwupdReleaseFlags flags = FWUPD_RELEASE_FLAG_NONE;
+		if (!fu_memcpy_safe((guint8 *)&flags,
+				    sizeof(flags),
+				    0x0, /* dst */
+				    g_bytes_get_data(blob, NULL),
+				    g_bytes_get_size(blob),
+				    0x0, /* src */
+				    sizeof(flags),
+				    error))
+			return FALSE;
+		if (flags & FWUPD_RELEASE_FLAG_TRUSTED_PAYLOAD)
+			fu_release_add_flag(self, FWUPD_RELEASE_FLAG_TRUSTED_PAYLOAD);
+		if (flags & FWUPD_RELEASE_FLAG_TRUSTED_METADATA)
+			fu_release_add_flag(self, FWUPD_RELEASE_FLAG_TRUSTED_METADATA);
+	}
+
+	/* do not require signatures for anything installed to the immutable datadir */
+	if (fu_release_get_flags(self) == FWUPD_RELEASE_FLAG_NONE && self->remote != NULL) {
+		g_debug("remote %s has kind=%s and so marking as trusted",
+			fwupd_remote_get_id(self->remote),
+			fwupd_remote_kind_to_string(fwupd_remote_get_kind(self->remote)));
+		if (fwupd_remote_get_kind(self->remote) == FWUPD_REMOTE_KIND_LOCAL ||
+		    fwupd_remote_get_kind(self->remote) == FWUPD_REMOTE_KIND_DIRECTORY) {
+			fu_release_add_flag(self, FWUPD_RELEASE_FLAG_TRUSTED_PAYLOAD);
+			fu_release_add_flag(self, FWUPD_RELEASE_FLAG_TRUSTED_METADATA);
+		} else {
+			fu_release_add_flag(self, FWUPD_RELEASE_FLAG_TRUSTED_METADATA);
+		}
+	}
+
+	/* success */
+	return TRUE;
 }
 
 /**
@@ -988,23 +1226,23 @@ fu_release_get_action_id(FuRelease *self)
 {
 	/* relax authentication checks for removable devices */
 	if (!fu_device_has_flag(self->device, FWUPD_DEVICE_FLAG_INTERNAL)) {
-		if (self->is_downgrade) {
-			if (self->trust_flags & FWUPD_TRUST_FLAG_PAYLOAD)
+		if (fu_release_has_flag(self, FWUPD_RELEASE_FLAG_IS_DOWNGRADE)) {
+			if (fu_release_has_flag(self, FWUPD_RELEASE_FLAG_TRUSTED_PAYLOAD))
 				return "org.freedesktop.fwupd.downgrade-hotplug-trusted";
 			return "org.freedesktop.fwupd.downgrade-hotplug";
 		}
-		if (self->trust_flags & FWUPD_TRUST_FLAG_PAYLOAD)
+		if (fu_release_has_flag(self, FWUPD_RELEASE_FLAG_TRUSTED_PAYLOAD))
 			return "org.freedesktop.fwupd.update-hotplug-trusted";
 		return "org.freedesktop.fwupd.update-hotplug";
 	}
 
 	/* internal device */
-	if (self->is_downgrade) {
-		if (self->trust_flags & FWUPD_TRUST_FLAG_PAYLOAD)
+	if (fu_release_has_flag(self, FWUPD_RELEASE_FLAG_IS_DOWNGRADE)) {
+		if (fu_release_has_flag(self, FWUPD_RELEASE_FLAG_TRUSTED_PAYLOAD))
 			return "org.freedesktop.fwupd.downgrade-internal-trusted";
 		return "org.freedesktop.fwupd.downgrade-internal";
 	}
-	if (self->trust_flags & FWUPD_TRUST_FLAG_PAYLOAD)
+	if (fu_release_has_flag(self, FWUPD_RELEASE_FLAG_TRUSTED_PAYLOAD))
 		return "org.freedesktop.fwupd.update-internal-trusted";
 	return "org.freedesktop.fwupd.update-internal";
 }
@@ -1014,7 +1252,7 @@ fu_release_get_action_id(FuRelease *self)
  * @release1: first task to compare.
  * @release2: second task to compare.
  *
- * Compares two install tasks.
+ * Compares two releases.
  *
  * Returns: 1, 0 or -1 if @release1 is greater, equal, or less than @release2, respectively.
  **/
@@ -1023,17 +1261,41 @@ fu_release_compare(FuRelease *release1, FuRelease *release2)
 {
 	FuDevice *device1 = fu_release_get_device(release1);
 	FuDevice *device2 = fu_release_get_device(release2);
-	if (fu_device_get_order(device1) < fu_device_get_order(device2))
+
+	/* device order, lower is better */
+	if (device1 != NULL && device2 != NULL && device1 != device2) {
+		if (fu_device_get_order(device1) < fu_device_get_order(device2))
+			return -1;
+		if (fu_device_get_order(device1) > fu_device_get_order(device2))
+			return 1;
+	}
+
+	/* release priority, higher is better */
+	if (release1->priority > release2->priority)
 		return -1;
-	if (fu_device_get_order(device1) > fu_device_get_order(device2))
+	if (release1->priority < release2->priority)
 		return 1;
-	return 0;
+
+	/* remote priority, higher is better */
+	if (release1->remote != NULL && release2->remote != NULL) {
+		if (fwupd_remote_get_priority(release1->remote) >
+		    fwupd_remote_get_priority(release2->remote))
+			return -1;
+		if (fwupd_remote_get_priority(release1->remote) <
+		    fwupd_remote_get_priority(release2->remote))
+			return 1;
+	}
+
+	/* FWUPD_DEVICE_FLAG_INSTALL_ALL_RELEASES has to be from oldest to newest */
+	return fu_version_compare(fu_release_get_version(release1),
+				  fu_release_get_version(release2),
+				  fu_device_get_version_format(device1));
 }
 
 static void
 fu_release_init(FuRelease *self)
 {
-	self->trust_flags = FWUPD_TRUST_FLAG_NONE;
+	fu_release_set_flags(self, FWUPD_RELEASE_FLAG_NONE);
 }
 
 static void
@@ -1042,6 +1304,7 @@ fu_release_finalize(GObject *obj)
 	FuRelease *self = FU_RELEASE(obj);
 
 	g_free(self->update_request_id);
+	g_free(self->device_version_old);
 	if (self->request != NULL)
 		g_object_unref(self->request);
 	if (self->device != NULL)
@@ -1050,8 +1313,8 @@ fu_release_finalize(GObject *obj)
 		g_object_unref(self->remote);
 	if (self->config != NULL)
 		g_object_unref(self->config);
-	if (self->blob_fw != NULL)
-		g_bytes_unref(self->blob_fw);
+	if (self->stream != NULL)
+		g_object_unref(self->stream);
 	if (self->soft_reqs != NULL)
 		g_ptr_array_unref(self->soft_reqs);
 	if (self->hard_reqs != NULL)
