@@ -1,7 +1,7 @@
 /*
- * Copyright (C) 2020 Richard Hughes <richard@hughsie.com>
+ * Copyright 2020 Richard Hughes <richard@hughsie.com>
  *
- * SPDX-License-Identifier: LGPL-2.1+
+ * SPDX-License-Identifier: LGPL-2.1-or-later
  */
 
 #include "config.h"
@@ -62,19 +62,19 @@ fu_security_attr_get_name(FwupdSecurityAttr *attr)
 		 * OTP = one time programmable */
 		return g_strdup(_("Intel BootGuard OTP fuse"));
 	}
-	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_INTEL_CET_ENABLED) == 0) {
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_CET_ENABLED) == 0) {
 		/* TRANSLATORS: Title: CET = Control-flow Enforcement Technology,
 		 * enabled means supported by the processor */
-		return g_strdup(_("Intel CET Enabled"));
+		return g_strdup(_("CET Platform"));
 	}
-	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_INTEL_CET_ACTIVE) == 0) {
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_CET_ACTIVE) == 0) {
 		/* TRANSLATORS: Title: CET = Control-flow Enforcement Technology,
-		 * active means being used by the OS */
-		return g_strdup(_("Intel CET Active"));
+		 * Utilized by OS means the distribution enabled it*/
+		return g_strdup(_("CET OS Support"));
 	}
-	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_INTEL_SMAP) == 0) {
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_SMAP) == 0) {
 		/* TRANSLATORS: Title: SMAP = Supervisor Mode Access Prevention */
-		return g_strdup(_("Intel SMAP"));
+		return g_strdup(_("SMAP"));
 	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_ENCRYPTED_RAM) == 0) {
 		/* TRANSLATORS: Title: Memory contents are encrypted, e.g. Intel TME */
@@ -113,6 +113,10 @@ fu_security_attr_get_name(FwupdSecurityAttr *attr)
 		/* TRANSLATORS: Title: SB is a way of locking down UEFI */
 		return g_strdup(_("UEFI secure boot"));
 	}
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_UEFI_BOOTSERVICE_VARS) == 0) {
+		/* TRANSLATORS: Title: Bootservice is when only readable from early-boot */
+		return g_strdup(_("UEFI bootservice variables"));
+	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_TPM_EMPTY_PCR) == 0) {
 		/* TRANSLATORS: Title: PCRs (Platform Configuration Registers) shouldn't be empty */
 		return g_strdup(_("TPM empty PCRs"));
@@ -146,6 +150,11 @@ fu_security_attr_get_name(FwupdSecurityAttr *attr)
 		 * end users on consumer boards */
 		return g_strdup(_("MEI override"));
 	}
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_MEI_KEY_MANIFEST) == 0) {
+		/* TRANSLATORS: Title: MEI = Intel Management Engine, and key refer
+		 * to the private/public key used to secure loading of firmware */
+		return g_strdup(_("MEI key manifest"));
+	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_MEI_VERSION) == 0) {
 		/* TRANSLATORS: Title: MEI = Intel Management Engine */
 		const gchar *kind = fwupd_security_attr_get_metadata(attr, "kind");
@@ -176,7 +185,7 @@ fu_security_attr_get_name(FwupdSecurityAttr *attr)
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_PLATFORM_DEBUG_ENABLED) == 0 ||
 	    g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_PLATFORM_DEBUG_LOCKED) == 0) {
 		/* TRANSLATORS: Title: Allows debugging of parts using proprietary hardware */
-		return g_strdup(_("Platform Debugging"));
+		return g_strdup(_("Platform debugging"));
 	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_SUPPORTED_CPU) == 0) {
 		/* TRANSLATORS: Title: if fwupd supports HSI on this chip */
@@ -184,7 +193,7 @@ fu_security_attr_get_name(FwupdSecurityAttr *attr)
 	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_AMD_ROLLBACK_PROTECTION) == 0) {
 		/* TRANSLATORS: Title: if firmware enforces rollback protection */
-		return g_strdup(_("Rollback protection"));
+		return g_strdup(_("Processor rollback protection"));
 	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_AMD_SPI_REPLAY_PROTECTION) == 0) {
 		/* TRANSLATORS: Title: if hardware enforces control of SPI replays */
@@ -202,7 +211,18 @@ fu_security_attr_get_name(FwupdSecurityAttr *attr)
 		/* TRANSLATORS: Title: if we are emulating a different host */
 		return g_strdup(_("Emulated host"));
 	}
-
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_BIOS_ROLLBACK_PROTECTION) == 0) {
+		/* TRANSLATORS: Title: if firmware enforces rollback protection */
+		return g_strdup(_("BIOS rollback protection"));
+	}
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_INTEL_GDS) == 0) {
+		/* TRANSLATORS: Title: GDS is where the CPU leaks information */
+		return g_strdup(_("Intel GDS mitigation"));
+	}
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_BIOS_CAPSULE_UPDATES) == 0) {
+		/* TRANSLATORS: Title: Whether BIOS Firmware updates is enabled */
+		return g_strdup(_("BIOS firmware updates"));
+	}
 	/* we should not get here */
 	return g_strdup(fwupd_security_attr_get_name(attr));
 }
@@ -254,14 +274,14 @@ fu_security_attr_get_title(FwupdSecurityAttr *attr)
 		/* TRANSLATORS: Title: BootGuard is a trademark from Intel */
 		return _("Intel BootGuard Fuse");
 	}
-	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_INTEL_CET_ENABLED) == 0 ||
-	    g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_INTEL_CET_ACTIVE) == 0) {
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_CET_ENABLED) == 0 ||
+	    g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_CET_ACTIVE) == 0) {
 		/* TRANSLATORS: Title: CET = Control-flow Enforcement Technology */
-		return _("Intel CET");
+		return _("Control-flow Enforcement Technology");
 	}
-	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_INTEL_SMAP) == 0) {
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_SMAP) == 0) {
 		/* TRANSLATORS: Title: SMAP = Supervisor Mode Access Prevention */
-		return _("Intel SMAP");
+		return _("Supervisor Mode Access Prevention");
 	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_ENCRYPTED_RAM) == 0) {
 		/* TRANSLATORS: Title: Memory contents are encrypted, e.g. Intel TME */
@@ -300,6 +320,10 @@ fu_security_attr_get_title(FwupdSecurityAttr *attr)
 		/* TRANSLATORS: Title: SB is a way of locking down UEFI */
 		return _("UEFI Secure Boot");
 	}
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_UEFI_BOOTSERVICE_VARS) == 0) {
+		/* TRANSLATORS: Title: Bootservice is when only readable from early-boot */
+		return _("UEFI Bootservice Variables");
+	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_TPM_EMPTY_PCR) == 0) {
 		/* TRANSLATORS: Title: PCRs (Platform Configuration Registers) shouldn't be empty */
 		return _("TPM Platform Configuration");
@@ -322,6 +346,11 @@ fu_security_attr_get_title(FwupdSecurityAttr *attr)
 		 * boards */
 		return _("Intel Management Engine Override");
 	}
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_MEI_KEY_MANIFEST) == 0) {
+		/* TRANSLATORS: Title: MEI = Intel Management Engine, and key refers
+		 * to the private/public key used to secure loading of firmware */
+		return _("MEI Key Manifest");
+	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_MEI_VERSION) == 0) {
 		/* TRANSLATORS: Title: MEI = Intel Management Engine */
 		return _("Intel Management Engine Version");
@@ -349,7 +378,7 @@ fu_security_attr_get_title(FwupdSecurityAttr *attr)
 	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_AMD_ROLLBACK_PROTECTION) == 0) {
 		/* TRANSLATORS: Title: if firmware enforces rollback protection */
-		return _("AMD Rollback Protection");
+		return _("AMD Secure Processor Rollback Protection");
 	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_AMD_SPI_REPLAY_PROTECTION) == 0) {
 		/* TRANSLATORS: Title: if hardware enforces control of SPI replays */
@@ -363,6 +392,18 @@ fu_security_attr_get_title(FwupdSecurityAttr *attr)
 		/* TRANSLATORS: Title: if the part has been fused */
 		return _("Fused Platform");
 	}
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_BIOS_ROLLBACK_PROTECTION) == 0) {
+		/* TRANSLATORS: Title: if firmware enforces rollback protection */
+		return _("BIOS Rollback Protection");
+	}
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_INTEL_GDS) == 0) {
+		/* TRANSLATORS: Title: GDS is where the CPU leaks information */
+		return _("Intel GDS Mitigation");
+	}
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_BIOS_CAPSULE_UPDATES) == 0) {
+		/* TRANSLATORS: Title: Whether BIOS Firmware updates is enabled */
+		return _("BIOS Firmware Updates");
+	}
 	return NULL;
 }
 
@@ -407,15 +448,15 @@ fu_security_attr_get_description(FwupdSecurityAttr *attr)
 		    "Intel BootGuard Error Policy ensures the device does not continue to start if "
 		    "its device software has been tampered with.");
 	}
-	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_INTEL_CET_ENABLED) == 0 ||
-	    g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_INTEL_CET_ACTIVE) == 0) {
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_CET_ENABLED) == 0 ||
+	    g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_CET_ACTIVE) == 0) {
 		/* TRANSLATORS: longer description */
-		return _("Intel Control-Flow Enforcement Technology detects and prevents certain "
+		return _("Control-Flow Enforcement Technology detects and prevents certain "
 			 "methods for running malicious software on the device.");
 	}
-	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_INTEL_SMAP) == 0) {
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_SMAP) == 0) {
 		/* TRANSLATORS: longer description */
-		return _("Intel Supervisor Mode Access Prevention ensures critical parts of "
+		return _("Supervisor Mode Access Prevention ensures critical parts of "
 			 "device memory are not accessed by less secure programs.");
 	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_ENCRYPTED_RAM) == 0) {
@@ -470,6 +511,10 @@ fu_security_attr_get_description(FwupdSecurityAttr *attr)
 		return _("UEFI Secure Boot prevents malicious software from being loaded when the "
 			 "device starts.");
 	}
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_UEFI_BOOTSERVICE_VARS) == 0) {
+		/* TRANSLATORS: longer description */
+		return _("UEFI boot service variables should not be readable from runtime mode.");
+	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_TPM_EMPTY_PCR) == 0) {
 		/* TRANSLATORS: longer description */
 		return _("The TPM (Trusted Platform Module) Platform Configuration is used to "
@@ -496,6 +541,11 @@ fu_security_attr_get_description(FwupdSecurityAttr *attr)
 		return _("Intel Management Engine Override disables checks for device software "
 			 "tampering.");
 	}
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_MEI_KEY_MANIFEST) == 0) {
+		/* TRANSLATORS: longer description */
+		return _("The Intel Management Engine Key Manifest must be valid so "
+			 "that the device firmware can be trusted by the CPU.");
+	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_MEI_VERSION) == 0) {
 		/* TRANSLATORS: longer description */
 		return _("The Intel Management Engine controls device components and needs "
@@ -527,299 +577,21 @@ fu_security_attr_get_description(FwupdSecurityAttr *attr)
 		return _("Each system should have tests to ensure firmware security.");
 	}
 	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_AMD_ROLLBACK_PROTECTION) == 0 ||
-	    g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_AMD_SPI_REPLAY_PROTECTION) == 0) {
+	    g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_AMD_SPI_REPLAY_PROTECTION) == 0 ||
+	    g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_BIOS_ROLLBACK_PROTECTION) == 0) {
 		/* TRANSLATORS: longer description */
-		return _("AMD Rollback Protection prevents device software from being downgraded "
+		return _("Rollback Protection prevents device software from being downgraded "
 			 "to an older version that has security problems.");
 	}
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_INTEL_GDS) == 0) {
+		/* TRANSLATORS: longer description */
+		return _("CPU Microcode must be updated to mitigate against various "
+			 "information-disclosure security issues.");
+	}
+	if (g_strcmp0(appstream_id, FWUPD_SECURITY_ATTR_ID_BIOS_CAPSULE_UPDATES) == 0) {
+		/* TRANSLATORS: longer description */
+		return _("Enabling firmware updates for the BIOS allows fixing security issues.");
+	}
+
 	return NULL;
 }
-
-const gchar *
-fu_security_attr_result_to_string(FwupdSecurityAttrResult result)
-{
-	if (result == FWUPD_SECURITY_ATTR_RESULT_VALID) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Valid");
-	}
-	if (result == FWUPD_SECURITY_ATTR_RESULT_NOT_VALID) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Invalid");
-	}
-	if (result == FWUPD_SECURITY_ATTR_RESULT_ENABLED) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Enabled");
-	}
-	if (result == FWUPD_SECURITY_ATTR_RESULT_NOT_ENABLED) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Disabled");
-	}
-	if (result == FWUPD_SECURITY_ATTR_RESULT_LOCKED) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Locked");
-	}
-	if (result == FWUPD_SECURITY_ATTR_RESULT_NOT_LOCKED) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Unlocked");
-	}
-	if (result == FWUPD_SECURITY_ATTR_RESULT_ENCRYPTED) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Encrypted");
-	}
-	if (result == FWUPD_SECURITY_ATTR_RESULT_NOT_ENCRYPTED) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Unencrypted");
-	}
-	if (result == FWUPD_SECURITY_ATTR_RESULT_TAINTED) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Tainted");
-	}
-	if (result == FWUPD_SECURITY_ATTR_RESULT_NOT_TAINTED) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Untainted");
-	}
-	if (result == FWUPD_SECURITY_ATTR_RESULT_FOUND) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Found");
-	}
-	if (result == FWUPD_SECURITY_ATTR_RESULT_NOT_FOUND) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Not found");
-	}
-	if (result == FWUPD_SECURITY_ATTR_RESULT_SUPPORTED) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Supported");
-	}
-	if (result == FWUPD_SECURITY_ATTR_RESULT_NOT_SUPPORTED) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("Not supported");
-	}
-	return NULL;
-}
-
-const gchar *
-fu_security_attr_get_result(FwupdSecurityAttr *attr)
-{
-	const gchar *tmp;
-
-	/* common case */
-	tmp = fu_security_attr_result_to_string(fwupd_security_attr_get_result(attr));
-	if (tmp != NULL)
-		return tmp;
-
-	/* fallback */
-	if (fwupd_security_attr_has_flag(attr, FWUPD_SECURITY_ATTR_FLAG_SUCCESS)) {
-		/* TRANSLATORS: Suffix: the HSI result */
-		return _("OK");
-	}
-
-	/* TRANSLATORS: Suffix: the fallback HSI result */
-	return _("Failed");
-}
-
-/**
- * fu_security_attrs_to_json_string：
- * Convert security attribute to JSON string.
- * @attrs: a pointer for a FuSecurityAttrs data structure.
- * @error: return location for an error
- *
- * fu_security_attrs_to_json_string() converts FuSecurityAttrs and return the
- * string pointer. The converted JSON format is shown as follows:
- * {
- *     "SecurityAttributes": [
- *         {
- *              "name": "aaa",
- *              "value": "bbb"
- *         }
- *     ]
- *  }
- *
- * Returns: A string and NULL on fail.
- *
- * Since: 1.7.0
- *
- */
-gchar *
-fu_security_attrs_to_json_string(FuSecurityAttrs *attrs, GError **error)
-{
-	g_autofree gchar *data = NULL;
-	g_autoptr(JsonGenerator) json_generator = NULL;
-	g_autoptr(JsonBuilder) builder = json_builder_new();
-	g_autoptr(JsonNode) json_root = NULL;
-	fu_security_attrs_to_json(attrs, builder);
-	json_root = json_builder_get_root(builder);
-	json_generator = json_generator_new();
-	json_generator_set_pretty(json_generator, TRUE);
-	json_generator_set_root(json_generator, json_root);
-	data = json_generator_to_data(json_generator, NULL);
-	if (data == NULL) {
-		g_set_error(error,
-			    FWUPD_ERROR,
-			    FWUPD_ERROR_INTERNAL,
-			    "Failed to convert security attribute to json.");
-		return NULL;
-	}
-	return g_steal_pointer(&data);
-}
-
-void
-fu_security_attrs_to_json(FuSecurityAttrs *attrs, JsonBuilder *builder)
-{
-	g_autoptr(GPtrArray) items = NULL;
-
-	json_builder_begin_object(builder);
-	json_builder_set_member_name(builder, "SecurityAttributes");
-	json_builder_begin_array(builder);
-	items = fu_security_attrs_get_all(attrs);
-	for (guint i = 0; i < items->len; i++) {
-		FwupdSecurityAttr *attr = g_ptr_array_index(items, i);
-		guint64 created = fwupd_security_attr_get_created(attr);
-		json_builder_begin_object(builder);
-		fwupd_security_attr_set_created(attr, 0);
-		fwupd_security_attr_to_json(attr, builder);
-		fwupd_security_attr_set_created(attr, created);
-		json_builder_end_object(builder);
-	}
-	json_builder_end_array(builder);
-	json_builder_end_object(builder);
-}
-
-gboolean
-fu_security_attrs_from_json(FuSecurityAttrs *attrs, JsonNode *json_node, GError **error)
-{
-	JsonArray *array;
-	JsonObject *obj;
-
-	/* sanity check */
-	if (!JSON_NODE_HOLDS_OBJECT(json_node)) {
-		g_set_error_literal(error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA, "not JSON object");
-		return FALSE;
-	}
-	obj = json_node_get_object(json_node);
-
-	/* this has to exist */
-	if (!json_object_has_member(obj, "SecurityAttributes")) {
-		g_set_error_literal(error,
-				    G_IO_ERROR,
-				    G_IO_ERROR_INVALID_DATA,
-				    "no SecurityAttributes property in object");
-		return FALSE;
-	}
-	array = json_object_get_array_member(obj, "SecurityAttributes");
-	for (guint i = 0; i < json_array_get_length(array); i++) {
-		JsonNode *node_tmp = json_array_get_element(array, i);
-		g_autoptr(FwupdSecurityAttr) attr = fwupd_security_attr_new(NULL);
-		if (!fwupd_security_attr_from_json(attr, node_tmp, error))
-			return FALSE;
-		fu_security_attrs_append(attrs, attr);
-	}
-
-	/* success */
-	return TRUE;
-}
-
-/**
- * fu_security_attrs_compare：
- * @attrs1: a #FuSecurityAttrs
- * @attrs2: another #FuSecurityAttrs, perhaps newer in some way
- *
- * Compares the two objects, returning the differences.
- *
- * If the two sets of attrs are considered the same then an empty array is returned.
- * Only the AppStream ID results are compared, extra metadata is ignored.
- *
- * Returns: (element-type FwupdSecurityAttr) (transfer container): differences
- */
-GPtrArray *
-fu_security_attrs_compare(FuSecurityAttrs *attrs1, FuSecurityAttrs *attrs2)
-{
-	g_autoptr(GHashTable) hash1 = g_hash_table_new(g_str_hash, g_str_equal);
-	g_autoptr(GHashTable) hash2 = g_hash_table_new(g_str_hash, g_str_equal);
-	g_autoptr(GPtrArray) array1 = fu_security_attrs_get_all(attrs1);
-	g_autoptr(GPtrArray) array2 = fu_security_attrs_get_all(attrs2);
-	g_autoptr(GPtrArray) results =
-	    g_ptr_array_new_with_free_func((GDestroyNotify)g_object_unref);
-
-	/* create hash tables of appstream-id -> FwupdSecurityAttr */
-	for (guint i = 0; i < array1->len; i++) {
-		FwupdSecurityAttr *attr1 = g_ptr_array_index(array1, i);
-		g_hash_table_insert(hash1,
-				    (gpointer)fwupd_security_attr_get_appstream_id(attr1),
-				    (gpointer)attr1);
-	}
-	for (guint i = 0; i < array2->len; i++) {
-		FwupdSecurityAttr *attr2 = g_ptr_array_index(array2, i);
-		g_hash_table_insert(hash2,
-				    (gpointer)fwupd_security_attr_get_appstream_id(attr2),
-				    (gpointer)attr2);
-	}
-
-	/* present in attrs2, not present in attrs1 */
-	for (guint i = 0; i < array2->len; i++) {
-		FwupdSecurityAttr *attr1;
-		FwupdSecurityAttr *attr2 = g_ptr_array_index(array2, i);
-		attr1 = g_hash_table_lookup(hash1, fwupd_security_attr_get_appstream_id(attr2));
-		if (attr1 == NULL) {
-			g_autoptr(FwupdSecurityAttr) attr = fwupd_security_attr_copy(attr2);
-			g_ptr_array_add(results, g_steal_pointer(&attr));
-			continue;
-		}
-	}
-
-	/* present in attrs1, not present in attrs2 */
-	for (guint i = 0; i < array1->len; i++) {
-		FwupdSecurityAttr *attr1 = g_ptr_array_index(array1, i);
-		FwupdSecurityAttr *attr2;
-		attr2 = g_hash_table_lookup(hash2, fwupd_security_attr_get_appstream_id(attr1));
-		if (attr2 == NULL) {
-			g_autoptr(FwupdSecurityAttr) attr = fwupd_security_attr_copy(attr1);
-			fwupd_security_attr_set_result(attr, FWUPD_SECURITY_ATTR_RESULT_UNKNOWN);
-			fwupd_security_attr_set_result_fallback(
-			    attr, /* flip these around */
-			    fwupd_security_attr_get_result(attr1));
-			g_ptr_array_add(results, g_steal_pointer(&attr));
-			continue;
-		}
-	}
-
-	/* find any attributes that differ */
-	for (guint i = 0; i < array2->len; i++) {
-		FwupdSecurityAttr *attr1;
-		FwupdSecurityAttr *attr2 = g_ptr_array_index(array2, i);
-		attr1 = g_hash_table_lookup(hash1, fwupd_security_attr_get_appstream_id(attr2));
-		if (attr1 == NULL)
-			continue;
-
-		/* result of specific attr differed */
-		if (fwupd_security_attr_get_result(attr1) !=
-		    fwupd_security_attr_get_result(attr2)) {
-			g_autoptr(FwupdSecurityAttr) attr = fwupd_security_attr_copy(attr1);
-			fwupd_security_attr_set_result(attr, fwupd_security_attr_get_result(attr2));
-			fwupd_security_attr_set_result_fallback(
-			    attr,
-			    fwupd_security_attr_get_result(attr1));
-			fwupd_security_attr_set_flags(attr, fwupd_security_attr_get_flags(attr2));
-			g_ptr_array_add(results, g_steal_pointer(&attr));
-		}
-	}
-
-	/* success */
-	return g_steal_pointer(&results);
-}
-
-/**
- * fu_security_attrs_equal：
- * @attrs1: a #FuSecurityAttrs
- * @attrs2: another #FuSecurityAttrs
- *
- * Tests the objects for equality. Only the AppStream ID results are compared, extra metadata
- * is ignored.
- *
- * Returns: %TRUE if the set of attrs can be considered equal
- */
-gboolean
-fu_security_attrs_equal(FuSecurityAttrs *attrs1, FuSecurityAttrs *attrs2)
-{
-	g_autoptr(GPtrArray) compare = fu_security_attrs_compare(attrs1, attrs2);
-	return compare->len == 0;
-}
